from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway


class LlmDecomposerGatewayAdapter:
    def __init__(self, llm_gateway: LlmGateway):
        self._llm_gateway = llm_gateway

    def generate_completion(self, user_message: str) -> str:
        return self._llm_gateway.generate_completion(user_message)



from typing import AsyncGenerator

try:
    from xai_sdk import Client
    from xai_sdk.chat import system, user
except ImportError:  # pragma: no cover - optional dependency
    Client = None
    system = None
    user = None
from analogai.deepthink.infrastructure.gateways.llm import config
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway


class XaiLlmGatewayImpl(LlmGateway):
    def __init__(
        self,
        system_prompt: str | None = None,
        max_tokens: int = 500,
        model: str | None = None,
    ):
        if Client is None or system is None or user is None:
            raise ImportError(
                "xai-sdk is required for the XAI gateway. Install it with 'pip install xai-sdk'."
            )
        api_key = config.XAI_API_KEY
        if not api_key:
            raise ValueError("XAI_API_KEY must be set for XAI gateway")
        if not model:
            raise ValueError("Model name must be provided for XAI gateway")
        self.model = model
        self.default_system_prompt = system_prompt or ""
        self.default_max_tokens = min(max_tokens, 500)
        self.client = Client(api_key=api_key)

    def set_model(self, model: str) -> None:
        self.model = model

    def set_system_prompt(self, system_prompt: str) -> None:
        self.default_system_prompt = system_prompt

    def generate_completion(
        self,
        user_message: str,
    ) -> str:
        final_system_prompt = self.default_system_prompt
        chat = self.client.chat.create(
            model=self.model,
            messages=[
                system(final_system_prompt),
                user(user_message),
            ],
        )
        response = chat.sample()
        content = getattr(response, "content", None)
        if content is None:
            return ""
        return str(content).strip()

    async def generate_streaming_completion(
        self,
        user_message: str,
    ) -> AsyncGenerator:
        output = self.generate_completion(user_message)
        if output:
            yield output

    def close(self) -> None:
        try:
            self.client.close()
        except AttributeError:
            pass


if __name__ == "__main__":
    gateway = XaiLlmGatewayImpl(model="grok-3")
    result = gateway.generate_completion("Hello, how are you?")
    print(result)


from analogai.foundation.common.design_patterns.state_machine import State
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.application.usecases.learning.make_direct_statement.utils.modifier_util import modifiers_match


class FindStatementState(State):
    def __init__(self, state_machine, statement_service):
        super().__init__(state_machine)
        self._statement_service = statement_service

    def execute(self):
        app_logger.info(f"Finding existing statement: subject={repr(self.context.subject_notion)}, complement={repr(self.context.complement_notion)}, relation={self.context.relation}")
        statement = self._find_statement()
        if statement:
            app_logger.info(f"Found existing statement: {repr(statement)}")
        else:
            app_logger.info("No existing statement found, will create new")
        return self._transition(statement)

    def _find_statement(self):
        relation = self.context.relation
        statement = self._statement_service.find(
            subject_notion=self.context.subject_notion,
            complement_notion=self.context.complement_notion,
            relation=relation,
            modifier=self.context.modifier,
        )
        if statement and not modifiers_match(statement.modifier, self.context.modifier):
            return None
        if statement:
            self.context.statement = statement
        return statement

    def _transition(self, statement):
        if not statement:
            return self._transition_to_create()
        return self._transition_to_check_update_applicable()

    def _transition_to_create(self):
        return self.state_machine.state_to_create_statement()

    def _transition_to_check_update_applicable(self):
        return self.state_machine.state_to_update_statement()



import unittest

from analogai.foundation.common.enums import RelationshipType
from analogai.foundation.core.value_objects.relation import Relation
from analogai.deepthink.presentation.intent_type import IntentType
from analogai.deepthink.presentation.parsers.parsers_factory import build_message_parser


class TestSentenceParserTransformed(unittest.TestCase):
    def setUp(self):
        self.parser = build_message_parser()

    def test_parses_statement(self):
        parsed_list = self.parser.parse('{"intent":"making_statement","subject":"socrates","relationship":"is_a","object":"mortal","probability":1.0}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(parsed.data.relation, Relation.from_type(RelationshipType.IS_A))
        self.assertEqual(parsed.data.subject.caption, "socrates")
        self.assertEqual(parsed.data.complement.caption, "mortal")

    def test_parses_price_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"iphone 12","attributes":[{"tag":"price","value":"800","unit":"usd"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "price")
        self.assertEqual(attr.value, 800.0)
        self.assertEqual(attr.unit, "USD")

    def test_parses_quantity_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"socrates","attributes":[{"tag":"height","value":"200","unit":"cm"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "height")
        self.assertEqual(attr.value, 200.0)
        self.assertEqual(attr.unit, "cm")

    def test_parses_weight_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"box","attributes":[{"tag":"weight","value":"10","unit":"kg"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "weight")
        self.assertEqual(attr.value, 10.0)
        self.assertEqual(attr.unit, "kg")

    def test_parses_width_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"table","attributes":[{"tag":"width","value":"120","unit":"cm"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "width")
        self.assertEqual(attr.value, 120.0)
        self.assertEqual(attr.unit, "cm")

    def test_parses_length_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"rope","attributes":[{"tag":"length","value":"25","unit":"m"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "length")
        self.assertEqual(attr.value, 25.0)
        self.assertEqual(attr.unit, "m")

    def test_parses_depth_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"lake","attributes":[{"tag":"depth","value":"30","unit":"m"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "depth")
        self.assertEqual(attr.value, 30.0)
        self.assertEqual(attr.unit, "m")

    def test_parses_area_statement(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"yard","attributes":[{"tag":"area","value":"50","unit":"sq ft"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "area")
        self.assertEqual(attr.value, 50.0)
        self.assertEqual(attr.unit, "sq ft")

    def test_parses_price_statement_price_tag(self):
        parsed_list = self.parser.parse('{"intent":"attribute_quantity_statement","subject":"iphone 12","attributes":[{"tag":"price","value":"850","unit":"usd"}]}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_STATEMENT)
        self.assertEqual(len(parsed.data.subject.attributes), 1)
        attr = parsed.data.subject.attributes[0]
        self.assertEqual(attr.tag, "price")
        self.assertEqual(attr.value, 850.0)
        self.assertEqual(attr.unit, "USD")

    def test_parses_modifier_tags(self):
        parsed_list = self.parser.parse('{"intent":"making_statement","time_from":"14:00","time_to":"16:00","subject":"it","relationship":"has_property","object":"rain","probability":1.0}')
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.data.from_time, "14:00")
        self.assertEqual(parsed.data.to_time, "16:00")

    def test_parses_multiple_sentences(self):
        parsed_list = self.parser.parse(
            '{"sentences":[{"intent":"making_statement","subject":"iphone 12","relationship":"is","object":"product","probability":1.0},{"intent":"making_statement","subject":"macbook pro","relationship":"is","object":"product","probability":1.0}]}',
        )
        self.assertEqual(len(parsed_list), 2)
        self.assertEqual(parsed_list[0].data.subject.caption, "iphone 12")
        self.assertEqual(parsed_list[0].data.complement.caption, "product")
        self.assertEqual(parsed_list[1].data.subject.caption, "macbook pro")
        self.assertEqual(parsed_list[1].data.complement.caption, "product")

    def test_transforms_condition_to_causal(self):
        json_input = '{"sentences":[{"intent":"making_statement","condition":{"subject":"it","relationship":"is","object":"day","location":"london"},"subject":"stock market","relationship":"work","probability":1.0}]}'
        parsed_list = self.parser.parse(json_input)
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.intent_type, IntentType.MAKING_CAUSAL_STATEMENT)
        self.assertEqual(parsed.data.relation, Relation.from_type(RelationshipType.CAUSES))
        self.assertIn("day", parsed.data.subject.caption)
        self.assertIn("london", parsed.data.subject.caption)
        self.assertEqual(parsed.data.complement.caption, "stock market work")

    def test_parses_until_time_as_to_time(self):
        parsed_list = self.parser.parse(
            '{"intent":"making_statement","time":"until 1846","subject":"arthur\'s magazine","relationship":"is","object":"active","probability":1.0}'
        )
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertIsNone(parsed.data.from_time)
        self.assertEqual(parsed.data.to_time, "1846")

    def test_parses_till_time_as_to_time(self):
        parsed_list = self.parser.parse(
            '{"intent":"making_statement","time":"till 1846","subject":"arthur\'s magazine","relationship":"is","object":"active","probability":1.0}'
        )
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertIsNone(parsed.data.from_time)
        self.assertEqual(parsed.data.to_time, "1846")

    def test_parses_after_time_as_from_time(self):
        parsed_list = self.parser.parse(
            '{"intent":"making_statement","time":"after 1846","subject":"arthur\'s magazine","relationship":"is","object":"active","probability":1.0}'
        )
        self.assertTrue(parsed_list)
        parsed = parsed_list[0]
        self.assertEqual(parsed.data.from_time, "1846")
        self.assertIsNone(parsed.data.to_time)


if __name__ == "__main__":
    unittest.main()




from __future__ import annotations

from typing import Optional

from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway
from analogai.deepthink.infrastructure.gateways.llm.llm_gateway_provider_factory import LlmGatewayProviderFactory
from analogai.deepthink.startup.extensions_service_factory import ExtensionsServiceFactory
from analogai.deepthink.features.question_answering.question_prompt_builder import QuestionPromptBuilder
from analogai.deepthink.features.question_answering.question_answering_service import QuestionAnsweringService
import analogai.deepthink.config as app_config


def create_question_answering_service(
    domain_service_factory,
    llm_gateway_instance: Optional[LlmGateway] = None,
    similarity_cutoff: Optional[float] = None,
) -> QuestionAnsweringService:
    if similarity_cutoff is None:
        similarity_cutoff = app_config.SIMILARITY_CUTOFF
    use_vector_search = app_config.USE_VECTOR_SEARCH
    belief_query_service = ExtensionsServiceFactory(domain_service_factory).get_belief_query_service()
    notion_service = domain_service_factory.get_notion_service()
    prompt_builder = QuestionPromptBuilder()
    system_prompt = prompt_builder.build_system_prompt()

    if llm_gateway_instance is not None:
        llm_gateway_instance.set_system_prompt(system_prompt)
        llm_gateway = llm_gateway_instance
    else:
        llm_gateway = LlmGatewayProviderFactory.create_gateway(
            system_prompt=system_prompt,
            max_tokens=300,
        )

    synthesizer_gateway = LlmGatewayProviderFactory.create_gateway(
        system_prompt="",
        max_tokens=200,
    )

    return QuestionAnsweringService(
        belief_query_service=belief_query_service,
        notion_service=notion_service,
        llm_gateway=llm_gateway,
        synthesizer_gateway=synthesizer_gateway,
        prompt_builder=prompt_builder,
        max_entity_depth=app_config.QUESTION_ANSWERING_MAX_ENTITY_DEPTH,
        similarity_cutoff=similarity_cutoff,
        use_vector_search=use_vector_search,
    )



from typing import Optional, Dict, Tuple

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_gateway_type import LlmGatewayType
# from analogai.deepthink.presentation.user_message import UserMessage
from analogai.deepthink.presentation.preprocessing.preprocessor import preprocess_text
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack


class Deepthinker:
    _DEFAULT_USER_ID = "default-user"
    _DEFAULT_AGENT_ID = "default-agent"

    def __init__(
        self,
        notion_enrichment_enabled: Optional[bool] = None,
        llm_gateway_instance: Optional[LlmGateway] = None,
        similarity_cutoff: Optional[float] = None,
        llm_provider: Optional[LlmProvider] = None,
        llm_model: Optional[str] = None,
    ):
        from analogai.deepthink.startup.bootstrap import create_pipeline
        from analogai.deepthink.config import (
            NOTION_ENRICHMENT_ENABLED,
            SIMILARITY_CUTOFF,
        )
        from analogai.deepthink.infrastructure.gateways.llm.llm_gateway_provider_factory import (
            LlmGatewayProviderFactory,
        )
        from analogai.deepthink.infrastructure.gateways.llm.llm_factory import LlmGatewayFactory
        from analogai.deepthink.startup.factories.domain_service_factory import DomainServiceFactory
        from analogai.deepthink.features.question_answering.factory import create_question_answering_service
        from analogai.deepthink.features.question_answering.question_router import QuestionRouter

        self._notion_enrichment_enabled = (
            NOTION_ENRICHMENT_ENABLED
            if notion_enrichment_enabled is None
            else notion_enrichment_enabled
        )
        self._similarity_cutoff = (
            SIMILARITY_CUTOFF if similarity_cutoff is None else similarity_cutoff
        )
        if not llm_model:
            raise ValueError("llm_model must be provided; defaults are disabled")
        if llm_provider is None:
            raise ValueError("llm_provider must be provided; defaults are disabled")

        self._llm_model = llm_model
        provider_map = {
            LlmProvider.ANTHROPIC: LlmGatewayType.ANTHROPIC,
            LlmProvider.OPENAI: LlmGatewayType.OPENAI,
            LlmProvider.AZURE: LlmGatewayType.AZURE,
            LlmProvider.AZURE_OPENAI: LlmGatewayType.AZURE_OPENAI,
            LlmProvider.XAI: LlmGatewayType.GROK,
            LlmProvider.BEDROCK: LlmGatewayType.BEDROCK,
        }
        self._llm_gateway_type = provider_map.get(llm_provider)
        if self._llm_gateway_type is None:
            raise ValueError(f"Unsupported llm_provider: {llm_provider}")
        self._llm_gateway_instance = llm_gateway_instance
        LlmGatewayProviderFactory.set_default_gateway_type(self._llm_gateway_type)
        LlmGatewayProviderFactory.set_default_model(self._llm_model)
        if self._llm_gateway_instance is not None:
            if hasattr(self._llm_gateway_instance, "set_model"):
                try:
                    self._llm_gateway_instance.set_model(self._llm_model)
                except Exception:
                    pass
            LlmGatewayFactory.set_default_gateway_instance(self._llm_gateway_instance)
        else:
            LlmGatewayFactory.set_default_gateway_type(self._llm_gateway_type)

        self._domain_service_factory = DomainServiceFactory()
        self._pipeline = create_pipeline(
            domain_service_factory=self._domain_service_factory,
            notion_enrichment_enabled=self._notion_enrichment_enabled,
            llm_gateway_instance=self._llm_gateway_instance,
        )
        self._question_answering_service = create_question_answering_service(
            self._domain_service_factory,
            llm_gateway_instance=self._llm_gateway_instance,
            similarity_cutoff=self._similarity_cutoff,
        )
        self._router = QuestionRouter(self._question_answering_service)
        self._internal_stacks: Dict[Tuple[str, str], PreviousMessagesStack] = {}

    def deepthink(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self.chat(
            text=text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def chat(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        if self._is_question(text, stack):
            answer = self.search(
                text=text,
                user_agent=resolved_user_agent,
                previous_messages_stack=stack,
            )
            return answer if answer else "I don't know"
        return self.remember(
            text=text,
            user_agent=resolved_user_agent,
            previous_messages_stack=stack,
        )

    def _is_question(
        self,
        text: str,
        stack: PreviousMessagesStack,
    ) -> bool:
        if self._router.is_question(text):
            return True
        if "?" in text:
            return True
        return False

    def remember(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        if text:
            text = text.replace('"', "")
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        stack.add_user(text)
        result = self._handle_statement(text, resolved_user_agent, stack)
        if result:
            stack.add_agent(result)
        return result

    def search(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        stack = self._resolve_prev_message_stack(previous_messages_stack, resolved_user_agent)
        stack.add_user(text)
        answer = self._question_answering_service.search(
            question_text=text,
            user_agent=resolved_user_agent,
        )
        if answer:
            stack.add_agent(answer)
        return answer

    def answer(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> Optional[str]:
        return self.search(
            text=text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def generate_context(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        resolved_user_agent = self._resolve_user_agent(user_agent)
        return self._question_answering_service.generate_context(
            question_text=text,
            user_agent=resolved_user_agent,
        )

    def get_context(
        self,
        text: str,
        user_agent: Optional[UserAgent] = None,
        previous_messages_stack: Optional[PreviousMessagesStack] = None,
    ) -> str:
        return self.generate_context(
            text=text,
            user_agent=user_agent,
            previous_messages_stack=previous_messages_stack,
        )

    def enrich_notions(self, notion: object) -> None:
        if not self._notion_enrichment_enabled:
            return
        from analogai.deepthink.features.notion_enrichment.registry import get_notion_enrichment_observer

        observer = get_notion_enrichment_observer()
        if observer is None:
            return
        observer.on_notion_created(notion)

    def _resolve_prev_message_stack(
        self,
        stack: Optional[PreviousMessagesStack],
        user_agent: UserAgent,
    ) -> PreviousMessagesStack:
        if stack is not None:
            return stack
        key = (user_agent.user_id, user_agent.agent_id)
        if key not in self._internal_stacks:
            self._internal_stacks[key] = PreviousMessagesStack()
        return self._internal_stacks[key]

    def _resolve_user_agent(
        self,
        user_agent: Optional[UserAgent],
    ) -> UserAgent:
        if user_agent is not None:
            return user_agent
        return UserAgent(
            user_id=self._DEFAULT_USER_ID,
            agent_id=self._DEFAULT_AGENT_ID,
            user_authority=1.0,
        )

    def _handle_statement(
        self,
        text: str,
        user_agent: UserAgent,
        stack: PreviousMessagesStack,
    ) -> Optional[str]:
        result = self._pipeline.execute(
            raw_message=text,
            user_agent=user_agent,
            previous_messages_stack=stack,
        )
        return result.message_text if result else None


DeepThinker = Deepthinker
Thinker = Deepthinker






import unittest
from unittest.mock import Mock
from analogai.foundation.common.enums import RelationshipType
from analogai.foundation.common.dtos.belief_expression import Relation, Modifier
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.time import Time
from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementRequest, MakeDirectStatementResponse
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.common.response_builder import LearnerResponseBuilder
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.context.learner_context import LearnerContext
from analogai.deepthink.application.usecases.learning.make_direct_statement.states.context.learner_context_handler import LearnerContextHandler
from analogai.deepthink.application.usecases.learning.make_direct_statement.usecase import MakeDirectStatementUseCase
from analogai.deepthink.tests.functional.application.usecases.learner.setup import LearnerTestSetup

class TestLearnerFunctional(unittest.TestCase):
    def setUp(self):
        self.setup = LearnerTestSetup()
        self.mock_output_port = self.setup.mock_output_port
        
        context = LearnerContext()
        context_handler = LearnerContextHandler(context, self.setup.notion_service)
        response_builder = LearnerResponseBuilder(context)
        
        self.usecase = MakeDirectStatementUseCase(
            statement_service=self.setup.statement_service,
            notion_service=self.setup.notion_service,
            output_port=self.mock_output_port,
            context_handler=context_handler,
            response_builder=response_builder
        )

    def tearDown(self):
        self.setup.clear()

    def test_learn_new_statement(self):
        request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="Dog"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        
        result = self.usecase.execute(request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertEqual(call_args.belief_expression.subject_notion_expression.caption, "Dog")
        self.assertEqual(call_args.belief_expression.complement_notion_expression.caption, "Animal")
        self.assertEqual(call_args.belief_expression.relation, Relation.from_type(RelationshipType.IS_A))
        self.assertEqual(call_args.belief_expression.probability, 0.8)

    def test_learn_statement_with_probability_zero_persists_negation(self):
        request = self.setup.create_learner_request(
            probability=0.0,
            subject_expression=NotionExpression(caption="Human"),
            object_expression=NotionExpression(caption="mortal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        result = self.usecase.execute(request)
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertEqual(call_args.belief_expression.probability, 0.0)
        self.assertIn("not", call_args.belief_expression.__str__().lower())

    def test_learn_statement_without_modifier(self):
        request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="Dog"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        self.assertIsNone(request.modifier)
        result = self.usecase.execute(request)
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertIsNone(call_args.belief_expression.modifier)

    def test_learn_statement_with_modifier(self):
        modifier = Modifier(location="Tbilisi")
        request = MakeDirectStatementRequest(
            user_agent=self.setup.user_agent,
            subject_notion_expression=NotionExpression(caption="it"),
            complement_notion_expression=NotionExpression(caption="rain"),
            relation=Relation.from_type(RelationshipType.DO),
            probability=1.0,
            modifier=modifier
        )
        result = self.usecase.execute(request)
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertIsNotNone(call_args.belief_expression.modifier)
        self.assertEqual(call_args.belief_expression.modifier.location, "Tbilisi")

    def test_learn_statement_with_modifier_time(self):
        from_time = Time(year=2024, month=12, day=26)
        modifier = Modifier(from_time=from_time)
        request = MakeDirectStatementRequest(
            user_agent=self.setup.user_agent,
            subject_notion_expression=NotionExpression(caption="it"),
            complement_notion_expression=NotionExpression(caption="rain"),
            relation=Relation.from_type(RelationshipType.DO),
            probability=1.0,
            modifier=modifier
        )
        result = self.usecase.execute(request)
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertIsNotNone(call_args.belief_expression.modifier)
        self.assertIsNotNone(call_args.belief_expression.modifier.from_time)
        self.assertEqual(call_args.belief_expression.modifier.from_time.year, 2024)

    def test_learn_duplicate_statement(self):
        initial_request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="Cat"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        self.usecase.execute(initial_request)
        self.mock_output_port.mock_output_create_statement_success.reset_mock()
        self.mock_output_port.mock_output_statement_already_known.reset_mock()
        
        request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="Cat"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        
        result = self.usecase.execute(request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_statement_already_known.assert_called_once()

    def test_learn_statement_with_different_probability(self):
        initial_request = self.setup.create_learner_request(
            probability=0.3,
            subject_expression=NotionExpression(caption="Bird"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        self.usecase.execute(initial_request)
        self.mock_output_port.mock_output_create_statement_success.reset_mock()
        self.mock_output_port.mock_output_statement_update_success.reset_mock()
        
        request = self.setup.create_learner_request(
            probability=0.9,
            subject_expression=NotionExpression(caption="Bird"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        
        result = self.usecase.execute(request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_statement_update_success.assert_called_once()

    def test_learn_statement_with_small_probability_difference(self):
        initial_request = self.setup.create_learner_request(
            probability=0.6,
            subject_expression=NotionExpression(caption="Fish"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        self.usecase.execute(initial_request)
        self.mock_output_port.mock_output_create_statement_success.reset_mock()
        self.mock_output_port.mock_output_statement_update_success.reset_mock()
        self.mock_output_port.mock_output_update_not_applicable.reset_mock()
        
        request = self.setup.create_learner_request(
            probability=0.75,
            subject_expression=NotionExpression(caption="Fish"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        
        result = self.usecase.execute(request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_update_not_applicable.assert_called_once()

    def test_learn_multiple_statements(self):
        learn_request = self.setup.create_learner_request(
            probability=0.7,
            subject_expression=NotionExpression(caption="Mammal"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        result1 = self.usecase.execute(learn_request)
        
        learn_request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="Reptile"),
            object_expression=NotionExpression(caption="Animal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        result2 = self.usecase.execute(learn_request)
        
        self.assertIsNotNone(result1)
        self.assertIsNotNone(result2)
        self.assertEqual(self.mock_output_port.mock_output_create_statement_success.call_count, 2)

    def test_learn_with_complex_relationship(self):
        learn_request = self.setup.create_learner_request(
            probability=0.85,
            subject_expression=NotionExpression(caption="Human"),
            object_expression=NotionExpression(caption="Mammal"),
            relation=Relation.from_type(RelationshipType.IS_A)
        )
        result = self.usecase.execute(learn_request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()
        call_args = self.mock_output_port.mock_output_create_statement_success.call_args[0][0]
        self.assertEqual(call_args.belief_expression.relation, Relation.from_type(RelationshipType.IS_A))
        self.assertEqual(call_args.belief_expression.probability, 0.85)

    def test_learn_statement_with_different_modifier_creates_new(self):
        from datetime import datetime, timedelta
        today = datetime.now()
        tomorrow = today + timedelta(days=1)
        today_modal = Modifier(from_time=Time(year=today.year, month=today.month, day=today.day))
        tomorrow_modal = Modifier(from_time=Time(year=tomorrow.year, month=tomorrow.month, day=tomorrow.day))
        
        initial_request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="it"),
            object_expression=NotionExpression(caption="rain"),
            relation=Relation.from_type(RelationshipType.DO)
        )
        initial_request.modifier = today_modal
        self.usecase.execute(initial_request)
        self.mock_output_port.mock_output_create_statement_success.reset_mock()
        
        request = self.setup.create_learner_request(
            probability=0.8,
            subject_expression=NotionExpression(caption="it"),
            object_expression=NotionExpression(caption="rain"),
            relation=Relation.from_type(RelationshipType.DO)
        )
        request.modifier = tomorrow_modal
        
        result = self.usecase.execute(request)
        
        self.assertIsNotNone(result)
        self.mock_output_port.mock_output_create_statement_success.assert_called_once()

if __name__ == '__main__':
    unittest.main()





from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.enums.certainty import Certainty
from analogai.deepthink.presentation.interpreters.interpreter import Interpreter
from analogai.deepthink.presentation.interpreters.interpreter_result import InterpreterResult
from analogai.deepthink.presentation.intent_type import IntentType
from analogai.deepthink.application.usecases.learning.make_conditional_statement.models import (
    CauseCriteria,
    MakeConditionalStatementRequest,
    CriterionData,
    CriterionValue,
)


class ConditionalStatementInterpreter(Interpreter):
    def interpret(self, parsed_message, user_agent, message_text):
        if parsed_message is None:
            return None
        if parsed_message.intent_type != IntentType.MAKING_CAUSAL_STATEMENT:
            return None
        
        data = parsed_message.data
        
        app_logger.debug(f"Causal interpreter - data.causes={data.causes}, data.effects={data.effects}")
        
        components = self._resolve_causal_components(data)
        if not components:
            return None

        cause_criteria_list, effect_criteria = components

        if not cause_criteria_list or not effect_criteria:
            return None
        
        probability = data.probability if data.probability is not None else 1.0
        
        request = MakeConditionalStatementRequest(
            user_agent=user_agent,
            probability=probability,
            causes=cause_criteria_list,
            effect=CauseCriteria(criteria=effect_criteria),
        )
        
        first_cause = cause_criteria_list[0].criteria
        app_logger.info(
            "Conditional statement interpreted: causes=%s, effect=%s %s %s in %s",
            len(cause_criteria_list),
            effect_criteria.subject_caption,
            effect_criteria.relation,
            effect_criteria.complement_caption,
            effect_criteria.location or ".",
        )
        
        return InterpreterResult(intent_type=IntentType.MAKING_CAUSAL_STATEMENT, request=request)

    def _resolve_causal_components(self, data):
        rules = (
            (self._has_causes_and_effects, self._from_causes_and_effects),
            (self._has_effects_only, self._from_effects_only),
            (self._always_true, self._from_data_only),
        )

        for predicate, builder in rules:
            if predicate(data):
                return builder(data)
        return None

    def _has_causes_and_effects(self, data) -> bool:
        return bool(data.causes and data.effects)

    def _has_effects_only(self, data) -> bool:
        return bool(data.effects)

    def _always_true(self, data) -> bool:
        return True

    def _from_causes_and_effects(self, data):
        effect_data = data.effects[0] if data.effects else None
        if not effect_data:
            return None
        cause_criteria_list = self._build_cause_criteria_list(data.causes)
        effect_criteria = self._build_criteria(effect_data)
        return cause_criteria_list, effect_criteria

    def _from_effects_only(self, data):
        effect_data = data.effects[0]
        cause_criteria_list = self._build_single_cause_list(data)
        effect_criteria = self._build_criteria(effect_data)
        return cause_criteria_list, effect_criteria

    def _from_data_only(self, data):
        cause_criteria_list = self._build_single_cause_list(data)
        effect_criteria = self._build_criteria(data, use_object_as_subject=True)
        return cause_criteria_list, effect_criteria

    def _build_cause_criteria_list(self, causes) -> list[CauseCriteria]:
        cause_criteria_list: list[CauseCriteria] = []
        for cause_data in causes or []:
            cause_criteria = self._build_criteria(cause_data)
            if cause_criteria:
                cause_criteria_list.append(CauseCriteria(criteria=cause_criteria))
        return cause_criteria_list

    def _build_single_cause_list(self, data) -> list[CauseCriteria]:
        cause_criteria = self._build_criteria(data)
        return [CauseCriteria(criteria=cause_criteria)] if cause_criteria else []
    
    def _extract_caption(self, notion_expression):
        if notion_expression is None:
            return None
        try:
            return notion_expression.caption
        except AttributeError:
            return str(notion_expression).strip()

    def _build_criteria(self, data, use_object_as_subject: bool = False) -> CriterionData | None:
        if data is None:
            return None

        subject_value = data.complement if use_object_as_subject else data.subject
        subject_caption = self._extract_caption(subject_value) or ""
        complement_caption = "" if use_object_as_subject else (self._extract_caption(data.complement) or "")

        relation = self._normalize_relation(data)

        location = data.location or ""
        time = data.time or ""
        from_time = data.from_time or ""
        to_time = data.to_time or ""
        quantity = data.quantity or ""
        deontic_modality = data.deontic_modality or ""
        certainty = data.certainty
        if certainty and not isinstance(certainty, Certainty):
            try:
                certainty = Certainty(str(certainty).lower())
            except ValueError:
                certainty = None
        attributes = data.attributes or []

        attribute_values = self._normalize_attributes(attributes)
        if not complement_caption and attribute_values:
            complement_caption = attribute_values[0].get("tag") or ""

        if not subject_caption:
            return None

        return CriterionData(
            subject_caption=subject_caption,
            complement_caption=complement_caption or subject_caption,
            relation=relation,
            location=location or None,
            time=time or None,
            from_time=from_time or None,
            to_time=to_time or None,
            quantity=quantity or None,
            deontic_modality=deontic_modality or None,
            certainty=certainty,
            attribute_values=attribute_values,
        )

    def _normalize_relation(self, data) -> str:
        relation = getattr(data, "relation", None) or "is"
        if not relation:
            relation = "is"

        if isinstance(relation, str):
            return relation

        for attr_name in ("value", "relation", "name"):
            if hasattr(relation, attr_name):
                attr_value = getattr(relation, attr_name)
                if isinstance(attr_value, str) and attr_value:
                    return attr_value
                if hasattr(attr_value, "value") and isinstance(attr_value.value, str):
                    return attr_value.value

        relationship_phrase = getattr(data, "relationship_phrase", None)
        if isinstance(relationship_phrase, str) and relationship_phrase:
            return relationship_phrase

        return str(relation)

    def _normalize_attributes(self, attributes) -> list[CriterionValue]:
        if not attributes:
            return []
        normalized: list[CriterionValue] = []
        for attr in attributes:
            tag = attr.tag
            values = attr.values or []
            value = values[0] if values else None
            unit = attr.unit
            min_value = attr.min_value
            max_value = attr.max_value
            if tag:
                normalized.append(
                    CriterionValue(
                        tag=tag,
                        value=value,
                        unit=unit,
                        min_value=min_value,
                        max_value=max_value,
                    )
                )
        return normalized



from typing import Optional

from pydantic import BaseModel, ConfigDict
from analogai.foundation.common.dtos.belief_expression import Modifier, Relation
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.dtos.belief_expression import BeliefExpression
from analogai.deepthink.application.value_objects.user_agent import UserAgent


class MakeDirectStatementRequest(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    subject_notion_expression: NotionExpression
    complement_notion_expression: NotionExpression
    relation: Relation
    probability: float
    modifier: Optional[Modifier] = None
    quantity: Optional[float] = None
    statement_name: Optional[str] = None
    complement_caption_rest: Optional[str] = None


class MakeDirectStatementResponse(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    belief_expression: BeliefExpression


MakeDirectStatementRequest.model_rebuild()
MakeDirectStatementResponse.model_rebuild()



from analogai.deepthink.config import PROBABILITY_THRESHOLD

def check_update_statement_applicable(existing_probability, proposed_probability, existing_validity, proposed_validity):
	probability_is_different = False
	has_enough_validity = False
	if abs(existing_probability - proposed_probability) > PROBABILITY_THRESHOLD:
		probability_is_different = True
	if proposed_validity >= existing_validity:
		has_enough_validity = True
	return probability_is_different and has_enough_validity






from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import MakeStatementOutputPort
from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementResponse


class CausalAlreadyKnownCapturePort(MakeStatementOutputPort):
    def __init__(self, inner: MakeStatementOutputPort):
        self._inner = inner
        self.already_known = False

    def output_create_statement_success(self, response: MakeDirectStatementResponse):
        self.already_known = False
        return self._inner.output_create_statement_success(response)

    def output_statement_already_known(self, response: MakeDirectStatementResponse):
        self.already_known = True
        return self._inner.output_statement_already_known(response)

    def output_statement_update_success(self, response: MakeDirectStatementResponse):
        self.already_known = False
        return self._inner.output_statement_update_success(response)

    def output_update_not_applicable(self, response: MakeDirectStatementResponse):
        self.already_known = False
        return self._inner.output_update_not_applicable(response)

    def output_forbidden(self, response: MakeDirectStatementResponse):
        self.already_known = False
        return self._inner.output_forbidden(response)




import unittest

from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.tests.functional.storage_cleaner import cleanup_storages
from analogai.deepthink.thinker import DeepThinker


class TestThinkerPreviousMessageStackIntegration(unittest.TestCase):
    def setUp(self):
        cleanup_storages()
        self._deepthink = DeepThinker()

    def tearDown(self):
        cleanup_storages()

    def test_conversation_saved_per_user_after_think(self):
        user_id = "user-a"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        self._deepthink.deepthink("socrates is human", user_id=user_id, agent_id=agent_id, previous_messages_stack=stack)
        history = list(stack.get_history())
        self.assertEqual(len(history), 1)
        self.assertIn("socrates", history[0].lower())
        self.assertIn("human", history[0].lower())

    def test_conversation_adapted_to_stack_and_used_for_pronoun_resolution(self):
        user_id = "user-b"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        self._deepthink.deepthink("john is mortal", user_id=user_id, agent_id=agent_id, previous_messages_stack=stack)
        result = self._deepthink.deepthink("he is philosopher", user_id=user_id, agent_id=agent_id, previous_messages_stack=stack)
        self.assertIsNotNone(result)
        self.assertIn("john", result.lower())
        self.assertNotIn(" he ", f" {result.lower()} ")
        history = list(stack.get_history())
        self.assertEqual(len(history), 2)

    def test_conversations_isolated_per_user(self):
        stack_user1 = PreviousMessagesStack()
        stack_user2 = PreviousMessagesStack()
        self._deepthink.deepthink("alice likes apples", user_id="user-1", agent_id="agent-1", previous_messages_stack=stack_user1)
        self._deepthink.deepthink("bob likes oranges", user_id="user-2", agent_id="agent-1", previous_messages_stack=stack_user2)
        history_user1 = list(stack_user1.get_history())
        history_user2 = list(stack_user2.get_history())
        self.assertEqual(history_user1, ["alice likes apples"])
        self.assertEqual(history_user2, ["bob likes oranges"])

    def test_conversation_adapted_to_stack_returns_persisted_history(self):
        user_id = "user-c"
        agent_id = "agent-1"
        stack = PreviousMessagesStack()
        self._deepthink.deepthink("maria reads books", user_id=user_id, agent_id=agent_id, previous_messages_stack=stack)
        history = list(stack.get_history())
        self.assertGreater(len(history), 0)
        self.assertIn("maria reads books", history)


if __name__ == "__main__":
    unittest.main()



from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.utils.statement_serializers.notion_serializer import (
    serialize_notion_expression,
)

from analogai.deepthink.application.usecases.learning.make_notion_statement.models import (
    MakeNotionStatementResponse,
)
from analogai.deepthink.application.usecases.learning.make_notion_statement.ports import (
    MakeNotionStatementOutputPort,
)
from analogai.deepthink.presentation.presenters.presenter_response import PresenterResponse


class MakeNotionStatementPresenter(MakeNotionStatementOutputPort):
    def output_notion_statement_saved(
        self,
        response: MakeNotionStatementResponse,
    ) -> PresenterResponse:
        app_logger.info("Outputting notion statement saved")
        message = serialize_notion_expression(response.notion_expression)
        return PresenterResponse(message=message)



def format_items_and(items: list[str]) -> str:
    if not items:
        return ""
    if len(items) == 1:
        return items[0]
    if len(items) == 2:
        return f"{items[0]} and {items[1]}"
    return ", ".join(items[:-1]) + f", and {items[-1]}"


def format_items_and_capitalized(items: list[str]) -> str:
    if not items:
        return ""
    if len(items) == 1:
        return items[0].capitalize()
    if len(items) == 2:
        return f"{items[0].capitalize()} and {items[1].capitalize()}"
    return ", ".join(n.capitalize() for n in items[:-1]) + f", and {items[-1].capitalize()}"



from typing import Optional

from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.utils.expression_factory import ExpressionFactory


class NotionExpressionBuilder:
    def __init__(self, expression_factory: Optional[ExpressionFactory] = None) -> None:
        self._expression_factory = expression_factory or ExpressionFactory()

    def build(self, caption: str, attributes=None) -> NotionExpression:
        return self._expression_factory.create_notion_expression(caption, attributes=attributes)



from analogai.deepthink.application.usecases.chat.chat_response_stack import ChatResponseStack
from analogai.deepthink.application.usecases.chat.shared import (
    RelationshipCategory,
)

__all__ = [
    'ChatResponseStack',
    'RelationshipCategory',
]



from enum import Enum


class LlmGatewayType(str, Enum):
    ANTHROPIC = "anthropic"
    OPENAI = "openai"
    AZURE = "azure"
    AZURE_OPENAI = "azure_openai"
    GROK = "grok"
    BEDROCK = "bedrock"



import os

from analogai.deepthink.thinker import DeepThinker
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_provider import LlmProvider


thinker = DeepThinker(
    llm_provider=LlmProvider.AZURE,
    llm_model="gpt-4o-mini",
)


if __name__ == "__main__":
    while True:
        message = input("Message: ")
        result = thinker.chat(message)
        print(result if result else "(no response)")



import re
from typing import List

class SentenceSplitter:
    _ABBREVIATIONS = {
        'mr', 'mrs', 'ms', 'dr', 'prof', 'sr', 'jr', 'vs', 'etc', 'inc', 'ltd',
        'corp', 'co', 'st', 'ave', 'blvd', 'rd', 'u.s', 'u.s.a', 'e.g', 'i.e',
        'a.m', 'p.m', 'ph.d', 'b.c', 'a.d', 'no', 'vol', 'pp', 'ed', 'est'
    }
    _ABBREVIATIONS_ALWAYS_MERGE = {
        'mr', 'mrs', 'ms', 'dr', 'prof', 'sr', 'jr'
    }
    _COMMA_PLACEHOLDER = "___COMMA___"
    _IF_COMMA_PATTERN = re.compile(r'\bif\b.*?,.*?', re.IGNORECASE)

    @classmethod
    def _is_likely_abbreviation(cls, word: str) -> bool:
        word_clean = word.lower().rstrip('.,!?;:')
        return word_clean in cls._ABBREVIATIONS

    @classmethod
    def split_into_sentences(cls, text: str) -> List[str]:
        text = text.strip()
        if not text:
            return []

        if cls._IF_COMMA_PATTERN.search(text):
            text = re.sub(r'(\bif\b[^,.?!]*),([^.?!]*)', r'\1' + cls._COMMA_PLACEHOLDER + r'\2', text, flags=re.IGNORECASE)

        sentences = []
        current_sentence = ""
        i = 0
        in_quotes = False
        
        while i < len(text):
            char = text[i]
            
            if char == '"':
                in_quotes = not in_quotes
                current_sentence += char
                if not in_quotes:
                    if i + 1 < len(text) and text[i + 1] == ' ':
                        if i + 2 < len(text):
                            next_word_start = i + 2
                            while next_word_start < len(text) and text[next_word_start] == ' ':
                                next_word_start += 1
                            if next_word_start < len(text) and text[next_word_start].isupper():
                                period_pos = current_sentence.rfind('.', 0, len(current_sentence) - 1)
                                if period_pos != -1:
                                    quote_pos = current_sentence.rfind('"', 0, period_pos)
                                    if quote_pos != -1:
                                        first_part = current_sentence[:len(current_sentence)].strip()
                                        if first_part:
                                            sentences.append(first_part)
                                        current_sentence = ""
                i += 1
                continue
            
            current_sentence += char
            
            if char in '.!?' and not in_quotes:
                if i + 1 < len(text):
                    next_char = text[i + 1]
                    if next_char == ' ':
                        if i + 2 < len(text):
                            next_word_start = i + 2
                            while next_word_start < len(text) and text[next_word_start] == ' ':
                                next_word_start += 1
                            
                            if next_word_start < len(text):
                                next_char_after_space = text[next_word_start]
                                
                                words = current_sentence.split()
                                if words:
                                    last_word = words[-1].rstrip('.,!?;:')
                                    if cls._is_likely_abbreviation(last_word):
                                        if last_word.lower() in cls._ABBREVIATIONS_ALWAYS_MERGE:
                                            i += 1
                                            continue
                                        elif next_char_after_space.islower():
                                            i += 1
                                            continue
                                
                                    first_word = words[0]
                                    if first_word and first_word[0].islower():
                                        i += 1
                                        continue
                                
                                if next_char_after_space.islower():
                                    i += 1
                                    continue
                                
                                if next_char_after_space.isupper():
                                    sentences.append(current_sentence.strip())
                                    current_sentence = ""
                                    i += 1
                                    continue
                    elif next_char.islower():
                        i += 1
                        continue
                    elif next_char in '.!?':
                        i += 1
                        continue
                else:
                    if current_sentence.strip():
                        sentences.append(current_sentence.strip())
                    current_sentence = ""
            
            i += 1
        
        if current_sentence.strip():
            sentences.append(current_sentence.strip())
        
        if not sentences:
            text = text.replace(cls._COMMA_PLACEHOLDER, ',')
            sentences = [text] if text else []
        
        filtered_result = []
        for s in sentences:
            s = s.replace(cls._COMMA_PLACEHOLDER, ',')
            s = s.strip()
            if not s:
                continue
            s_no_punct = re.sub(r'[.!?;:,\s"\']+', '', s)
            if len(s_no_punct) <= 1:
                continue
            if re.match(r'^[.!?;:,\s"\']+$', s):
                continue
            filtered_result.append(s)
        
        return filtered_result

def split_into_sentences(text: str) -> List[str]:
    return SentenceSplitter.split_into_sentences(text)

if __name__ == "__main__":
    test_text = "I'm good, aren't I? If you go, I'll stay. You're mortal. Socrates isn't human!"
    print(split_into_sentences(test_text))





from dataclasses import dataclass
from typing import Any

from analogai.deepthink.presentation.intent_type import IntentType


@dataclass
class InterpreterResult:
    intent_type: IntentType
    request: Any



from typing import Optional, Any
from analogai.deepthink.presentation.intents.intent import Intent
from analogai.deepthink.presentation.presenters.presenter_response import PresenterResponse
from analogai.foundation.common.logging.app_logger import app_logger


class MakeConditionalStatementIntent(Intent):
    def __init__(self, controller):
        self.controller = controller

    def execute(self, request: Any) -> Optional[PresenterResponse]:
        app_logger.important("Processing make conditional statement intent")
        if request is None:
            app_logger.warning("[MakeConditionalStatementIntent] request is None")
            return None
        return self.controller.learn_conditional_statement(request)



from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.foundation.common.design_patterns.state_machine import State
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.enums import RelationshipType
from analogai.foundation.core.value_objects import Relation
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.application.utils.relationship_phrase_util import is_semantic_relationship
from .common.response_builder import LearnerResponseBuilder
from analogai.deepthink.application.usecases.learning.make_direct_statement.ports import MakeStatementOutputPort


class CreateStatementState(State):
    def __init__(
        self,
        state_machine,
        statement_service,
        notion_service: NotionService,
        output_port: MakeStatementOutputPort,
        response_builder: LearnerResponseBuilder,
    ):
        super().__init__(state_machine)
        self._statement_service = statement_service
        self._notion_service = notion_service
        self._response_builder = response_builder
        self._output_port = output_port

    def execute(self):
        app_logger.info(f"Creating new statement: subject={repr(self.context.subject_notion)}, complement={repr(self.context.complement_notion)}, relation={self.context.relation}")
        
        if self._is_subject_object_same():
            app_logger.info("Forbidden: subject and object are the same")
            return self._output_forbidden_response()

        if not is_semantic_relationship(self.context.relation):
            app_logger.info("Non-semantic relationship, creating simple statement")
            return self._create_statement_and_output()

        app_logger.info("Semantic relationship, generating statement hierarchy")
        main_statement = self._generate_statement_hierarchy()
        if main_statement is None:
            app_logger.info("Forbidden: statement hierarchy generation failed")
            return self._output_forbidden_response()

        self._store_statement_in_context(main_statement)
        app_logger.info(f"Successfully created statement: {repr(main_statement)}")
        return self._build_and_output_response()

    def _create_statement_and_output(self):
        relation = self.context.relation
        if self.context.relation and self.context.relation.is_of_type(RelationshipType.DO):
            complement_caption_rest = self.context.complement_caption_rest or self.context.complement_notion_expression.caption
        else:
            complement_caption_rest = self.context.complement_notion_expression.caption
        complement_expression = NotionExpression(caption=complement_caption_rest) if complement_caption_rest else self.context.complement_notion_expression
        complement_notion = self._notion_service.find_or_create(
            self.context.user_agent.user_id,
            self.context.user_agent.agent_id,
            complement_expression,
        )
        self._notify_notion_created(complement_notion)
        if self._is_notions_same(self.context.subject_notion, complement_notion):
            return self._output_forbidden_response()
        self._statement_service.create(
            user_id=self.context.user_agent.user_id,
            agent_id=self.context.user_agent.agent_id,
            subject_notion=self.context.subject_notion,
            relation=relation,
            complement_notion=complement_notion,
            probability=self.context.proposed_probability,
            validity=self.context.proposed_validity,
            modifier=self.context.modifier,
        )
        return self._build_and_output_response()

    def _generate_statement_hierarchy(self):
        subject_expression = self.context.subject_notion_expression
        complement_expression = self.context.complement_notion_expression

        if subject_expression and subject_expression.children:
            statements = []
            for child_expression in subject_expression.children:
                child_notion = self._notion_service.find_or_create(
                    self.context.user_agent.user_id,
                    self.context.user_agent.agent_id,
                    child_expression
                )
                self._notify_notion_created(child_notion)
                if self._is_notions_same(child_notion, self.context.complement_notion):
                    return None
                statement = self._create_statement_for_pair(child_notion, self.context.complement_notion)
                statements.append(statement)
            return statements[0] if statements else None

        if complement_expression and complement_expression.children:
            statements = []
            for child_expression in complement_expression.children:
                child_notion = self._notion_service.find_or_create(
                    self.context.user_agent.user_id,
                    self.context.user_agent.agent_id,
                    child_expression
                )
                self._notify_notion_created(child_notion)
                if self._is_notions_same(self.context.subject_notion, child_notion):
                    return None
                statement = self._create_statement_for_pair(self.context.subject_notion, child_notion)
                statements.append(statement)
            return statements[0] if statements else None

        return self._create_statement_for_pair(self.context.subject_notion, self.context.complement_notion)

    def _store_statement_in_context(self, statement):
        self.context.statement = statement

    def _create_statement_for_pair(self, subject_notion, complement_notion):
        if self._is_notions_same(subject_notion, complement_notion):
            return None
        relation = self.context.relation
        return self._statement_service.create(
            user_id=self.context.user_agent.user_id,
            agent_id=self.context.user_agent.agent_id,
            subject_notion=subject_notion,
            relation=relation,
            complement_notion=complement_notion,
            probability=self.context.proposed_probability,
            validity=self.context.proposed_validity,
            modifier=self.context.modifier,
        )

    def _is_subject_object_same(self):
        return self._is_notions_same(self.context.subject_notion, self.context.complement_notion)

    def _is_notions_same(self, subject_notion, complement_notion):
        if subject_notion is None or complement_notion is None:
            return False
        return subject_notion.id == complement_notion.id

    def _notify_notion_created(self, notion) -> None:
        if notion is None:
            return
        try:
            from analogai.deepthink.config import NOTION_ENRICHMENT_ENABLED
            if not NOTION_ENRICHMENT_ENABLED:
                return
            from analogai.deepthink.features.notion_enrichment.registry import get_notion_enrichment_observer
            observer = get_notion_enrichment_observer()
            if observer is None:
                return
            observer.on_notion_created(notion)
        except Exception:
            return

    def _output_forbidden_response(self):
        response = self._response_builder.build_response()
        return self._output_port.output_forbidden(response)

    def _build_and_output_response(self):
        response = self._response_builder.build_response()
        return self._output_port.output_create_statement_success(response)




from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.deepthink.features.notion_enrichment.enricher_service import NotionEnricherService
from analogai.deepthink.features.notion_enrichment.observer import NotionEnrichmentObserver
from analogai.deepthink.features.notion_enrichment.ports import NotionEnrichmentGatewayPort
from analogai.deepthink.features.notion_enrichment.registry import register_notion_enrichment_observer


def create_notion_enrichment_observer(
    statement_service,
    notion_service: NotionService,
    enrichment_gateway: NotionEnrichmentGatewayPort,
) -> NotionEnrichmentObserver:
    enricher_service = NotionEnricherService(
        statement_service=statement_service,
        notion_service=notion_service,
        gateway=enrichment_gateway,
    )
    return NotionEnrichmentObserver(enricher_service=enricher_service)


def create_and_register_notion_enrichment_observer(
    statement_service,
    notion_service: NotionService,
    enrichment_gateway: NotionEnrichmentGatewayPort,
) -> None:
    observer = create_notion_enrichment_observer(
        statement_service=statement_service,
        notion_service=notion_service,
        enrichment_gateway=enrichment_gateway,
    )
    register_notion_enrichment_observer(observer)




import json
import boto3
from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway
from analogai.deepthink.infrastructure.gateways.llm import config


class BedrockLlmGatewayImpl(LlmGateway):
    def __init__(self, system_prompt: str | None = None, max_tokens: int = 4096, model: str | None = None):
        if not config.AWS_ACCESS_KEY_ID or not config.AWS_SECRET_ACCESS_KEY:
            raise ValueError("AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be set for Bedrock gateway")
        if not model:
            raise ValueError("Model name must be provided for Bedrock gateway")
        self.model = model
        self.default_system_prompt = system_prompt or ""
        self.default_max_tokens = min(max_tokens, 4096)
        self.client = boto3.client(
            service_name="bedrock-runtime",
            region_name="us-east-1",
            aws_access_key_id=config.AWS_ACCESS_KEY_ID,
            aws_secret_access_key=config.AWS_SECRET_ACCESS_KEY,
        )

    def set_model(self, model_id: str) -> None:
        self.model = model_id

    def set_system_prompt(self, system_prompt: str) -> None:
        self.default_system_prompt = system_prompt

    def generate_completion(
        self,
        user_message: str,
    ) -> str:
        final_system_prompt = self.default_system_prompt
        max_tokens = self.default_max_tokens

        request_body = {
            "anthropic_version": "bedrock-2023-05-31",
            "max_tokens": max_tokens,
            "messages": [
                {"role": "user", "content": user_message}
            ],
            "system": final_system_prompt,
        }

        response = self.client.invoke_model(
            modelId=self.model,
            body=json.dumps(request_body),
        )
        response_body = json.loads(response.get("body").read())
        return response_body["content"][0]["text"]

    async def generate_streaming_completion(
        self,
        user_message: str,
    ):
        final_system_prompt = self.default_system_prompt
        max_tokens = self.default_max_tokens

        request_body = {
            "anthropic_version": "bedrock-2023-05-31",
            "max_tokens": max_tokens,
            "messages": [
                {"role": "user", "content": user_message}
            ],
            "system": final_system_prompt,
        }

        stream = self.client.invoke_model_with_response_stream(
            modelId=self.model,
            body=json.dumps(request_body),
        )
        for event in stream:
            if "chunk" in event and "bytes" in event["chunk"]:
                chunk_data = json.loads(event["chunk"]["bytes"].decode())
                if "delta" in chunk_data and "text" in chunk_data["delta"]:
                    yield event


if __name__ == "__main__":
    gateway = BedrockLlmGatewayImpl(model="anthropic.claude-3-sonnet-20240229-v1:0")
    result = gateway.generate_completion("Hello, how are you?")
    print(result)


from __future__ import annotations

from analogai.foundation.common.logging.app_logger import app_logger


def is_question(text: str) -> bool:
    if not text:
        return False
    trimmed = text.strip()
    if "?" in trimmed:
        app_logger.info(
            f"[QuestionDetector] is_question=True (reason=question_mark) text={trimmed!r}"
        )
        return True
    app_logger.info(
        f"[QuestionDetector] is_question=False (reason=no_question_mark) text={trimmed!r}"
    )
    return False



IRREGULAR_PAST = {
    "sell": "sold",
    "buy": "bought",
    "make": "made",
    "do": "did",
    "have": "had",
    "go": "went",
    "get": "got",
    "take": "took",
    "give": "gave",
    "see": "saw",
    "come": "came",
    "think": "thought",
    "find": "found",
    "tell": "told",
    "know": "knew",
    "become": "became",
    "leave": "left",
    "put": "put",
    "mean": "meant",
    "keep": "kept",
    "let": "let",
    "begin": "began",
    "seem": "seemed",
    "help": "helped",
    "show": "showed",
    "hear": "heard",
    "play": "played",
    "run": "ran",
    "move": "moved",
    "live": "lived",
    "believe": "believed",
    "hold": "held",
    "bring": "brought",
    "happen": "happened",
    "write": "wrote",
    "provide": "provided",
    "sit": "sat",
    "stand": "stood",
    "lose": "lost",
    "pay": "paid",
    "meet": "met",
    "include": "included",
    "continue": "continued",
    "set": "set",
    "learn": "learned",
    "change": "changed",
    "lead": "led",
    "understand": "understood",
    "watch": "watched",
    "follow": "followed",
    "stop": "stopped",
    "create": "created",
    "speak": "spoke",
    "read": "read",
    "allow": "allowed",
    "add": "added",
    "spend": "spent",
    "grow": "grew",
    "open": "opened",
    "walk": "walked",
    "win": "won",
    "offer": "offered",
    "remember": "remembered",
    "love": "loved",
    "consider": "considered",
    "appear": "appeared",
    "wait": "waited",
    "serve": "served",
    "die": "died",
    "send": "sent",
    "expect": "expected",
    "build": "built",
    "stay": "stayed",
    "fall": "fell",
    "cut": "cut",
    "reach": "reached",
    "kill": "killed",
    "remain": "remained",
    "suggest": "suggested",
    "raise": "raised",
    "pass": "passed",
}


def to_past_tense(verb: str) -> str:
    if not verb:
        return ""
    lower = verb.lower()
    if lower in IRREGULAR_PAST:
        return IRREGULAR_PAST[lower]
    return lower + "d" if lower.endswith("e") else lower + "ed"


def pluralize_category(category_caption: str | None) -> str:
    if not category_caption:
        return ""
    if category_caption.endswith("s"):
        return category_caption
    return category_caption + "s"


def get_plural_relationship(relation: str) -> str:
    relation_value = relation

    if not relation_value:
        relation_value = "related to"

    try:
        from analogai.foundation.common.utils.statement_serializer.relation_verbalization_map import (
            RELATION_TO_VERBALIZATION,
        )
        from analogai.foundation.common.utils.statement_serializer import Verbalization
    except ModuleNotFoundError:
        try:
            from analogai.foundation.common.utils.statement_serializers.relation_verbalization_map import (
                RELATION_TO_VERBALIZATION,
            )
            from analogai.foundation.common.utils.statement_serializers import Verbalization
        except ModuleNotFoundError:
            return str(relation_value)

    verbalization = RELATION_TO_VERBALIZATION.get(relation_value)
    if verbalization is None:
        verbalization = Verbalization.RELATED_TO
    forms = verbalization._forms(is_plural=True)
    return forms["positive"]



from unittest.mock import Mock
from analogai.deepthink.tests.mocks.entities.mock_notion_factory import MockNotionFactory
from analogai.foundation.common.enums import RelationshipType
from .mock_user_factory import MockUserFactory

class MockStatementFactory:
    @staticmethod
    def create(content='default_statement', subject_notion=None, complement_notion=None, relationship_type=RelationshipType.IS_A, probability=0.3, validity=0.4, author=None, agent=None):
        if subject_notion is None:
            subject_notion = MockNotionFactory.create()
        if complement_notion is None:
            complement_notion = MockNotionFactory.create()
        if agent is None:
            agent = Mock()
        statement = Mock(name=f"{subject_notion.caption}->{relationship_type.value}->{complement_notion.caption}")
        statement.content = content
        statement.subject_notion = subject_notion
        statement.complement_notion = complement_notion
        statement.relation = relationship_type.value
        statement.probability = float(probability)
        statement.validity = float(validity)
        statement.author = author or MockUserFactory.create()
        statement.agent = agent
        return statement 





from __future__ import annotations

from analogai.deepthink.infrastructure.gateways.llm.gateway import LlmGateway


class QuestionLlmCaller:
    def __init__(self, gateway: LlmGateway):
        self._gateway = gateway

    def answer(self, user_message: str) -> str:
        return self._gateway.generate_completion(user_message)



from typing import List, Dict, Any, Optional, Protocol, runtime_checkable

from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.extensions.belief_query.services.belief_query_service import BeliefQueryService


class TraversableNotion:
    def __init__(self, notion: Notion, learnings_service: BeliefQueryService):
        self._notion = notion
        self._learnings_service = learnings_service

    def query_in(self, **criteria: Any) -> List[Belief]:
        builder = self._learnings_service.search(
            self._notion.user_id, self._notion.agent_id
        ).where(Criterion.ComplementCaption, self._notion.caption)
        if criteria:
            criterion_map = self._to_criterion_map(criteria)
            builder = builder.apply(criterion_map)
        return builder.list()

    def get_learnings_out(self, **criteria: Any) -> List[Belief]:
        builder = self._learnings_service.search(
            self._notion.user_id, self._notion.agent_id
        ).where(Criterion.SubjectCaption, self._notion.caption)
        if criteria:
            criterion_map = self._to_criterion_map(criteria)
            builder = builder.apply(criterion_map)
        return builder.list()

    def get_beliefs_in(self, relation: Relation, modifier: Optional[Modifier] = None) -> List[Belief]:
        criteria = self._build_belief_criteria(relation, modifier)
        builder = self._learnings_service.search(
            self._notion.user_id, self._notion.agent_id
        ).where(Criterion.ComplementCaption, self._notion.caption)
        if criteria:
            builder = builder.apply(criteria)
        return builder.list()

    def query_out(self, relation: Relation, modifier: Optional[Modifier] = None) -> List[Belief]:
        criteria = self._build_belief_criteria(relation, modifier)
        builder = self._learnings_service.search(
            self._notion.user_id, self._notion.agent_id
        ).where(Criterion.SubjectCaption, self._notion.caption)
        if criteria:
            builder = builder.apply(criteria)
        return builder.list()

    def _build_belief_criteria(
        self, relation: Relation, modifier: Optional[Modifier]
    ) -> Dict[Criterion, Any]:
        if not isinstance(relation, Relation):
            raise TypeError("relation must be a Relation instance")
        criteria: Dict[Criterion, Any] = {
            Criterion.Relation: relation,
        }
        if modifier is None:
            return criteria
        if modifier.location is not None:
            criteria[Criterion.Location] = modifier.location
        if modifier.deontic_modality is not None:
            criteria[Criterion.DeonticModality] = modifier.deontic_modality
        if modifier.from_time is not None:
            criteria[Criterion.FromTime] = modifier.from_time
        if modifier.to_time is not None:
            criteria[Criterion.ToTime] = modifier.to_time
        if modifier.attributes:
            criteria[Criterion.AttributeValue] = modifier.attributes
        return criteria

    def _to_criterion_map(self, criteria: Dict[str, Any]) -> Dict[Criterion, Any]:
        result = {}
        for k, v in criteria.items():
            try:
                result[Criterion(k)] = v
            except ValueError:
                pass
        return result

    @property
    def id(self) -> str:
        return self._notion.id

    @property
    def caption(self) -> str:
        return self._notion.caption

    @property
    def user_id(self) -> str:
        return self._notion.user_id

    @property
    def agent_id(self) -> str:
        return self._notion.agent_id

    @property
    def notion(self) -> Notion:
        return self._notion


def get_traversable_notion(
    belief_query_service: BeliefQueryService,
    notion: Notion,
) -> TraversableNotion:
    if isinstance(belief_query_service, TraversableNotionProvider):
        traversable = belief_query_service.get_traversable_notion(notion)
        if traversable is not None:
            return traversable
    return TraversableNotion(notion, belief_query_service)


@runtime_checkable
class TraversableNotionProvider(Protocol):
    def get_traversable_notion(self, notion: Notion) -> Optional[TraversableNotion]:
        ...



                                                          
                                                                                                 
import sys
from pathlib import Path

                                         
parent_dir = Path(__file__).parent.parent

                                       
import importlib.util
config_file = parent_dir / 'config.py'
spec = importlib.util.spec_from_file_location("application.config_module", str(config_file))
config_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(config_module)

                                             
__all__ = []
for attr_name in dir(config_module):
    if not attr_name.startswith('_'):
        globals()[attr_name] = vars(config_module)[attr_name]
        __all__.append(attr_name)






import unittest
from analogai.deepthink.presentation.postprocessing.postprocessors.pronoun_postprocessior import (
    postprocesses_pronouns,
    AgentMessagePronounPostprocessor,
    PlaceholderPronounReplacer,
    _lowercase_first_letter,
    _uppercase_first_letter,
)
from analogai.deepthink.presentation.enums.pronoun_placeholder import PronounPlaceholder


class TestPronounPostprocessorTemplate(unittest.TestCase):
    def test_lowercase_first_letter_empty(self):
        self.assertEqual(_lowercase_first_letter(""), "")
        self.assertEqual(_lowercase_first_letter("   "), "   ")

    def test_lowercase_first_letter_single_word(self):
        self.assertEqual(_lowercase_first_letter("Hello"), "hello")

    def test_lowercase_first_letter_leading_space(self):
        self.assertEqual(_lowercase_first_letter("  Hello world"), "  hello world")

    def test_uppercase_first_letter_single_word(self):
        self.assertEqual(_uppercase_first_letter("hello"), "Hello")

    def test_uppercase_first_letter_leading_space(self):
        self.assertEqual(_uppercase_first_letter("  hello world"), "  Hello world")


class TestPronounPostprocessorToPronoun(unittest.TestCase):
    def test_agent_placeholder_to_i(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{agent} think so")
        self.assertEqual(result, "I think so")

    def test_agent_placeholder_is_to_i_am(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{agent} is great")
        self.assertEqual(result, "I am great")

    def test_respondent_placeholder_to_you(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{respondent} are here")
        self.assertEqual(result, "You are here")

    def test_respondent_placeholder_is_to_you_are(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{respondent} is smart")
        self.assertEqual(result, "You are smart")

    def test_respondent_placeholder_with_identifier(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{respondent} is founder")
        self.assertEqual(result, "You are founder")

    def test_company_placeholder_to_we(self):
        company = PronounPlaceholder.COMPANY_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{company} sell iphones")
        self.assertEqual(result, "We sell iphones")

    def test_company_placeholder_is_to_we_are(self):
        company = PronounPlaceholder.COMPANY_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{company} is ready")
        self.assertEqual(result, "We are ready")

    def test_i_does_to_i_do(self):
        result = postprocesses_pronouns("I does think so")
        self.assertEqual(result, "I do think so")

    def test_you_does_to_you_do(self):
        result = postprocesses_pronouns("you does know")
        self.assertEqual(result, "You do know")


class TestPronounPostprocessorFirstLetterCase(unittest.TestCase):
    def test_uppercase_restored_after_replace(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{agent} is happy")
        self.assertEqual(result[0], "I")
        self.assertEqual(result, "I am happy")

    def test_leading_space_preserves_uppercase_restore(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"  {agent} is here")
        self.assertEqual(result, "  I am here")


class TestPronounPostprocessorNoDuplicateReplacement(unittest.TestCase):
    def test_single_pass_agent(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(agent)
        self.assertEqual(result, "I")

    def test_agent_placeholder_s_to_my(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{agent}'s voice")
        self.assertEqual(result, "My voice")

    def test_respondent_placeholder_s_to_your(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        result = postprocesses_pronouns(f"{respondent}'s name")
        self.assertEqual(result, "Your name")



from unittest.mock import Mock
from analogai.foundation.modules.user.user import User


class _statementServiceTestAdapter:
    def __init__(self, statement_service):
        self._service = statement_service

    def create(self, user_id, agent_id, subject_notion, complement_notion, relationship_type,
               probability=1.0, validity=1.0, modifier=None, **kwargs):
        relation = relationship_type.value
        return self._service.create(
            user_id=user_id,
            agent_id=agent_id,
            subject_notion=subject_notion,
            relation=relation,
            complement_notion=complement_notion,
            probability=probability,
            validity=validity,
            modifier=modifier,
            **{k: v for k, v in kwargs.items() if k in ('quantity', 'id')}
        )

    def update_probability(self, statement, probability):
        target = getattr(statement, "statement_for_update", lambda: statement)()
        return self._service.update_probability(target, probability)

    def update_validity(self, statement, validity):
        return self._service.update_validity(statement, validity)
from analogai.foundation.modules.agent.agent import Agent
from analogai.deepthink.startup.factories.domain_service_factory import DomainServiceFactory
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.deepthink.application.usecases.chat.chat_response_stack import ChatResponseStack
from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.tests.functional.storage_cleaner import cleanup_storages


class ChatterTestSetup:
    def __init__(self):
        domain_service_factory = DomainServiceFactory()
        self.domain_service_factory = domain_service_factory

        self.notion_service = domain_service_factory.get_notion_service()
        self.belief_service = domain_service_factory.get_belief_service()
        self.statement_service = domain_service_factory.get_statement_service()
        self.statement_service = _statementServiceTestAdapter(self.statement_service)
        self.deductive_inference_service = domain_service_factory.get_deductive_inference_service()
        self.belief_query_service = self._get_belief_query_service(domain_service_factory)
        
        self.user = User(id="test-user", name="Test User", email="test@example.com")
        self.agent = Agent(id="test-agent", knowledge_base="Test KB", creator=self.user)
        self.user_agent = UserAgent(
            user_id=self.user.id,
            agent_id=self.agent.id,
            user_authority=1.0
        )
        self.mock_output_port = Mock()
        self.response_stack = ChatResponseStack()
        
        self.socrates_notion = self.create_notion("socrates")
        self.human_notion = self.create_notion("human")
        self.mortal_notion = self.create_notion("mortal")

    def _get_belief_query_service(self, domain_service_factory):
        from analogai.deepthink.startup.extensions_service_factory import ExtensionsServiceFactory
        return ExtensionsServiceFactory(domain_service_factory).get_belief_query_service()

    def create_notion(self, caption):
        return self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            NotionExpression(caption=caption)
        )

    def tear_down(self):
        cleanup_storages()





from typing import Optional, Callable

from analogai.deepthink.presentation.intent_type import IntentType
from analogai.deepthink.presentation.intents.intent import Intent, NullIntent
from analogai.deepthink.presentation.intents.collection import (
    MakeStatementIntent,
    MakeConditionalStatementIntent,
    MakeNotionAttributeStatementIntent,
)
from analogai.deepthink.presentation.controllers.learning_controller import LearningController
from analogai.deepthink.startup.factories.domain_service_factory import DomainServiceFactory


class IntentsFactory:
    def __init__(
        self,
        domain_service_factory: DomainServiceFactory,
        learning_controller: LearningController,
        aura_service: Optional[object] = None,
    ):
        self.domain_service_factory = domain_service_factory
        self.learning_controller = learning_controller
        self.aura_service = aura_service
        self._null_intent = NullIntent()
        self._intent_builders = self._build_intent_registry()

    def _build_intent_registry(self) -> dict[IntentType, Callable[[], Intent]]:
        learning_controller = self.learning_controller

        return {
            IntentType.MAKING_STATEMENT: lambda: MakeStatementIntent(learning_controller),
            IntentType.MAKING_CAUSAL_STATEMENT: lambda: MakeConditionalStatementIntent(learning_controller),
            IntentType.MAKING_NOTION_ATTRIBUTE_STATEMENT: lambda: MakeNotionAttributeStatementIntent(learning_controller),
        }

    def create_intents(self) -> dict[IntentType, Intent]:
        return {
            intent_type: builder()
            for intent_type, builder in self._intent_builders.items()
        }

    def get_intent(self, intent_type: Optional[IntentType]) -> Intent:
        if intent_type is None:
            return self._null_intent
        builder = self._intent_builders.get(intent_type)
        if builder is None:
            return self._null_intent
        return builder()





from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.common.dtos.notion_expression import NotionExpression


def generate_expression(notion: Notion):
    attributes = getattr(notion, "attributes", None)
    return NotionExpression(caption=notion.caption, attributes=attributes)





import unittest

from analogai.deepthink.tests.e2e.base_deepthink_e2e_test import BaseThinkerE2ETest


class TestCausalHypotheticalThinkerE2E(BaseThinkerE2ETest):

    def test_conditional_statement_with_time_modifiers(self):
        response = self._think("if today it rains, tomorrow there will be floods")
        self.assertIsNotNone(response)
        self.assertIn("if", response.lower())
        self.assertIn("then", response.lower())

    def test_conditional_statement_with_location_modifiers(self):
        response = self._think("if there is heavy rain in the mountains, there is flooding in the valley")
        self.assertIsNotNone(response)
        self.assertIn("if", response.lower())
        self.assertIn("then", response.lower())



if __name__ == "__main__":
    unittest.main()



from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field
from analogai.foundation.common.enums.certainty import Certainty
from analogai.foundation.common.dtos.belief_expression import BeliefExpression
from analogai.deepthink.application.value_objects.user_agent import UserAgent


class CriterionValue(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    tag: str
    value: Optional[float] = None
    unit: Optional[str] = None
    min_value: Optional[float] = None
    max_value: Optional[float] = None


class CriterionData(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    subject_caption: str
    complement_caption: Optional[str] = None
    relation: Optional[str] = None
    location: Optional[str] = None
    time: Optional[str] = None
    from_time: Optional[str] = None
    to_time: Optional[str] = None
    quantity: Optional[float] = None
    deontic_modality: Optional[str] = None
    certainty: Optional[Certainty] = None
    attribute_values: list[CriterionValue] = Field(default_factory=list)


class CauseCriteria(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    criteria: CriterionData


class MakeConditionalStatementRequest(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    probability: float
    causes: list[CauseCriteria]
    effect: CauseCriteria


class MakeConditionalStatementResponse(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")

    user_agent: UserAgent
    belief_expression: BeliefExpression
    causes: list[CauseCriteria]
    effect: CauseCriteria
    probability: float
    validity: float = 1.0
    effect_belief_id: Optional[str] = None


MakeConditionalStatementRequest.model_rebuild()
MakeConditionalStatementResponse.model_rebuild()



import os
import sys
from importlib.util import find_spec
from pathlib import Path
from dotenv import load_dotenv
from analogai.deepthink.infrastructure.gateways.llm.enums.llm_gateway_type import LlmGatewayType

def _find_project_root(start: Path) -> Path | None:
    current = start
    while True:
        if (current / "pyproject.toml").exists() or (current / ".env").exists():
            return current
        if current.parent == current:
            return None
        current = current.parent


def _find_module_root(module_name: str) -> Path | None:
    spec = find_spec(module_name)
    if spec and spec.origin:
        return _find_project_root(Path(spec.origin).resolve())
    return None


def _ordered_env_roots() -> list[Path]:
    roots: list[Path] = []
    env_file = os.getenv("ENV_FILE")
    if env_file:
        env_path = Path(env_file).expanduser().resolve()
        roots.append(env_path.parent)

    argv0 = Path(sys.argv[0]).expanduser().resolve() if sys.argv and sys.argv[0] else None
    if argv0 and argv0.exists():
        entry_root = _find_project_root(argv0.parent)
        if entry_root:
            roots.append(entry_root)

    deepthink_root = _find_module_root("deepthink")
    if deepthink_root:
        roots.append(deepthink_root)

    analogai_root = _find_module_root("analogai")
    if analogai_root:
        roots.append(analogai_root)

    seen: set[Path] = set()
    ordered: list[Path] = []
    for root in roots:
        if root not in seen:
            ordered.append(root)
            seen.add(root)
    return ordered


def _load_envs_in_order() -> None:
    for root in _ordered_env_roots():
        env_path = root / ".env"
        if env_path.exists():
            load_dotenv(env_path, override=False)


_load_envs_in_order()

ENVIRONMENT = os.getenv('ENVIRONMENT', 'development')

if ENVIRONMENT not in ['development', 'production']:
    raise ValueError(f"Environment variable ENVIRONMENT must be one of: development or production")

QUICK_TESTING = os.getenv('QUICK_TESTING', 'false').lower() == 'true'
FINAL_APPLICATION_TESTING = os.getenv('FINAL_APPLICATION_TESTING', 'false').lower() == 'true'
if FINAL_APPLICATION_TESTING:
    QUICK_TESTING = False

DEBUG = os.getenv('DEBUG', 'false').lower() == 'true'

if QUICK_TESTING:
    LLM_PROCESSING_ENABLED = False
    NOTION_ENRICHMENT_ENABLED = False
    # NOTION_ENRICHMENT_ENABLED = False
    # STATEMENT_DECOMPOSITION_ENABLED = False
    print(f"QUICK_TESTING mode enabled - LLM_PROCESSING_ENABLED={LLM_PROCESSING_ENABLED}")
else:
    LLM_PROCESSING_ENABLED = os.getenv('LLM_PROCESSING_ENABLED', 'true').lower() == 'true'
    NOTION_ENRICHMENT_ENABLED = os.getenv('NOTION_ENRICHMENT_ENABLED', 'true').lower() == 'true'

    # NOTION_ENRICHMENT_ENABLED = os.getenv('NOTION_ENRICHMENT_ENABLED', 'true').lower() == 'true'
    # STATEMENT_DECOMPOSITION_ENABLED = os.getenv('STATEMENT_DECOMPOSITION_ENABLED', 'false').lower() == 'true'
    print(f"QUICK_TESTING mode disabled LLM_PROCESSING_ENABLED={LLM_PROCESSING_ENABLED}")

TIME_SENSITIVE_TRANSFORMATIONS_ENABLED = os.getenv('TIME_SENSITIVE_TRANSFORMATIONS_ENABLED', 'false').lower() == 'true'

PROBABILITY_THRESHOLD = 0.4

ENRICHMENT_LIMIT = 10


def _parse_int(value: str | None, default: int) -> int:
    try:
        return int(str(value).strip())
    except Exception:
        return default


def _parse_optional_float(value: str | None, default: float | None = None) -> float | None:
    if value is None:
        return default
    text = str(value).strip()
    if not text:
        return default
    try:
        return float(text)
    except Exception:
        return default


def _parse_bool(value: str | None, default: bool) -> bool:
    if value is None:
        return default
    text = str(value).strip().lower()
    if text in {"1", "true", "yes", "y", "on"}:
        return True
    if text in {"0", "false", "no", "n", "off"}:
        return False
    return default


QUESTION_ANSWERING_MAX_BELIEFS = _parse_int(
    os.getenv("QUESTION_ANSWERING_MAX_BELIEFS", "12"),
    12,
)

QUESTION_ANSWERING_MAX_ENTITY_DEPTH = _parse_int(
    os.getenv("QUESTION_ANSWERING_MAX_ENTITY_DEPTH", "3"),
    3,
)

USE_VECTOR_SEARCH = _parse_bool(os.getenv("USE_VECTOR_SEARCH"), True)
SIMILARITY_CUTOFF = _parse_optional_float(os.getenv("SIMILARITY_CUTOFF"), 0.4)


def _resolve_llm_gateway_type(value: str) -> LlmGatewayType:
    normalized = (value or "").strip().lower().replace("-", "_")
    mapping = {
        "anthropic": LlmGatewayType.ANTHROPIC,
        "claude": LlmGatewayType.ANTHROPIC,
        "openai": LlmGatewayType.OPENAI,
        "azure": LlmGatewayType.AZURE,
        "azure_openai": LlmGatewayType.AZURE_OPENAI,
        "azureopenai": LlmGatewayType.AZURE_OPENAI,
        "grok": LlmGatewayType.GROK,
        "xai": LlmGatewayType.GROK,
        "grok3": LlmGatewayType.GROK,
        "grok4_fast_non_reasoning": LlmGatewayType.GROK,
        "bedrock": LlmGatewayType.BEDROCK,

        # Backward-compatible keys mapped to Azure
        "azure_gpt4o_mini": LlmGatewayType.AZURE,
        "azure_gpt4o": LlmGatewayType.AZURE,
        "azure_gpt4_1_nano": LlmGatewayType.AZURE,
        "azure_openai_gpt5_1_chat": LlmGatewayType.AZURE_OPENAI,
        "azure_openai_gpt5_2_chat": LlmGatewayType.AZURE_OPENAI,
        "azure_openai_codex_mini": LlmGatewayType.AZURE_OPENAI,
        "azure_deepseek_r1": LlmGatewayType.AZURE,
        "azure_deepseek_v3_2": LlmGatewayType.AZURE,
        "gpt4o_mini": LlmGatewayType.AZURE,
        "gpt4o": LlmGatewayType.AZURE,
        "gpt4_1_nano": LlmGatewayType.AZURE,
        "gpt5_1_chat": LlmGatewayType.AZURE,
        "gpt5_2_chat": LlmGatewayType.AZURE,
        "codex_mini": LlmGatewayType.AZURE,
        "deepseek_r1": LlmGatewayType.AZURE,
        "deepseek_v3_2": LlmGatewayType.AZURE,
    }
    if normalized in mapping:
        return mapping[normalized]
    raise ValueError(
        "Unknown DEFAULT_LLM_GATEWAY_TYPE value: "
        f"'{value}'. Please provide a valid gateway type."
    )


_DEFAULT_LLM_GATEWAY_TYPE_VALUE = os.getenv("DEFAULT_LLM_GATEWAY_TYPE")
DEFAULT_LLM_GATEWAY_TYPE = (
    _resolve_llm_gateway_type(_DEFAULT_LLM_GATEWAY_TYPE_VALUE)
    if _DEFAULT_LLM_GATEWAY_TYPE_VALUE
    else None
)




# STATEMENT_DECOMPOSITION_ENABLED = os.getenv('STATEMENT_DECOMPOSITION_ENABLED', 'true').lower() == 'true'

print(f"Environment is set to {ENVIRONMENT} at Thinker")



"""
Comparison strategies for different types of text comparisons.
Follows the Strategy pattern and Single Responsibility Principle.
"""
from abc import ABC, abstractmethod
from typing import Any


def _normalize_for_comparison(value: Any) -> str:
    if value is None:
        return ""
    return str(value).strip().lower()


class ComparisonStrategy(ABC):
    """Abstract base class for comparison strategies."""
    
    @abstractmethod
    def compare(self, value1: Any, value2: Any) -> bool:
        """Compare two values according to the specific strategy.
        
        Args:
            value1: First value to compare
            value2: Second value to compare
            
        Returns:
            True if values match according to the strategy, False otherwise
        """
        pass


class ExactComparisonStrategy(ComparisonStrategy):
    """Compares values exactly after normalization."""
    
    def compare(self, value1: Any, value2: Any) -> bool:
        """Compare two values for exact equality after normalization.
        
        Args:
            value1: First value to compare
            value2: Second value to compare
            
        Returns:
            True if normalized values are exactly equal
        """
        return _normalize_for_comparison(value1) == _normalize_for_comparison(value2)


class ContainsComparisonStrategy(ComparisonStrategy):
    """Checks if one value contains another after normalization."""
    
    def compare(self, value1: Any, value2: Any) -> bool:
        """Check if value1 contains value2 after normalization.
        
        Args:
            value1: The container value
            value2: The value to search for
            
        Returns:
            True if normalized value1 contains normalized value2
        """
        norm1 = _normalize_for_comparison(value1)
        norm2 = _normalize_for_comparison(value2)
        return norm2 in norm1


class BidirectionalContainsStrategy(ComparisonStrategy):
    """Checks if either value contains the other after normalization."""
    
    def compare(self, value1: Any, value2: Any) -> bool:
        """Check if either value contains the other after normalization.
        
        Args:
            value1: First value
            value2: Second value
            
        Returns:
            True if value1 contains value2 OR value2 contains value1
        """
        norm1 = _normalize_for_comparison(value1)
        norm2 = _normalize_for_comparison(value2)
        return norm2 in norm1 or norm1 in norm2


class TextComparator:
    """
    Context class that uses a comparison strategy.
    Follows the Strategy pattern for flexible comparison behavior.
    """
    
    def __init__(self, strategy: ComparisonStrategy):
        """Initialize the comparator with a specific strategy.
        
        Args:
            strategy: The comparison strategy to use
        """
        self._strategy = strategy
    
    def compare(self, value1: Any, value2: Any) -> bool:
        """Compare two values using the configured strategy.
        
        Args:
            value1: First value to compare
            value2: Second value to compare
            
        Returns:
            Comparison result according to the strategy
        """
        return self._strategy.compare(value1, value2)
    
    def set_strategy(self, strategy: ComparisonStrategy):
        """Change the comparison strategy.
        
        Args:
            strategy: The new comparison strategy to use
        """
        self._strategy = strategy


# Convenience functions for common comparison operations
def are_equal_normalized(value1: Any, value2: Any) -> bool:
    """
    Check if two values are equal after normalization.
    
    Args:
        value1: First value
        value2: Second value
        
    Returns:
        True if values are equal after normalization
    """
    strategy = ExactComparisonStrategy()
    return strategy.compare(value1, value2)


def contains_normalized(container: Any, value: Any) -> bool:
    """
    Check if container contains value after normalization.
    
    Args:
        container: The container value
        value: The value to search for
        
    Returns:
        True if container contains value after normalization
    """
    strategy = ContainsComparisonStrategy()
    return strategy.compare(container, value)


def either_contains_normalized(value1: Any, value2: Any) -> bool:
    """
    Check if either value contains the other after normalization.
    
    Args:
        value1: First value
        value2: Second value
        
    Returns:
        True if either value contains the other
    """
    strategy = BidirectionalContainsStrategy()
    return strategy.compare(value1, value2)



from typing import Optional, List
from pydantic import BaseModel, ConfigDict, Field
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects import Relation
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.deepthink.application.value_objects.user_agent import UserAgent
from enum import Enum


class Attribute(BaseModel):
    tag: str
    values: List[float] = Field(default_factory=list)
    unit: Optional[str] = None
    min_value: Optional[float] = None
    max_value: Optional[float] = None


class StatementType(str, Enum):
    SIMPLE = "simple"
    ATTRIBUTE_QUANTITY = "attribute_quantity"
    NOTION_ATTRIBUTE = "notion_attribute"


class LearnStatementRequest(BaseModel):
    model_config = ConfigDict(strict=True, extra="forbid")
    
    user_agent: UserAgent
    statement_type: StatementType
    subject: NotionExpression
    complement: Optional[NotionExpression] = None
    relation: Optional[Relation] = None
    probability: Optional[float] = None
    modifier: Optional[Modifier] = None
    attributes: List[Attribute] = Field(default_factory=list)
    quantity: Optional[float] = None
    from_time: Optional[str] = None
    to_time: Optional[str] = None
    location: Optional[str] = None
    deontic_modality: Optional[str] = None



from __future__ import annotations

from typing import Any, Optional


def extract_text_from_response(data: dict) -> str:
    if not isinstance(data, dict):
        return ""

    if "output" in data:
        output = data["output"]
        if isinstance(output, str):
            return output.strip()
        if isinstance(output, list) and output:
            first = output[0]
            if isinstance(first, dict):
                content = first.get("content") or first.get("text")
                if isinstance(content, str):
                    return content.strip()
            return str(first).strip()
        if isinstance(output, dict):
            content = output.get("content")
            if isinstance(content, str):
                return content.strip()
            delta = output.get("delta")
            if isinstance(delta, dict):
                delta_content = delta.get("content")
                if isinstance(delta_content, str):
                    return delta_content.strip()
        return str(output).strip()

    if "choices" in data and isinstance(data.get("choices"), list) and data["choices"]:
        first_choice = data["choices"][0]
        if isinstance(first_choice, dict):
            message = first_choice.get("message")
            if isinstance(message, dict):
                content = message.get("content")
                if isinstance(content, str):
                    return content.strip()
            content = first_choice.get("content")
            if isinstance(content, str):
                return content.strip()

    if "content" in data and isinstance(data.get("content"), str):
        return str(data["content"]).strip()

    return ""


def extract_stream_delta(data: dict) -> Optional[str]:
    if not isinstance(data, dict):
        return None

    if "output" in data:
        output = data["output"]
        if isinstance(output, str):
            return output
        if isinstance(output, list) and output:
            first = output[0]
            if isinstance(first, dict):
                content = first.get("content") or first.get("text")
                if isinstance(content, str):
                    return content
                delta = first.get("delta")
                if isinstance(delta, dict):
                    delta_content = delta.get("content")
                    if isinstance(delta_content, str):
                        return delta_content
            return None
        if isinstance(output, dict):
            delta = output.get("delta")
            if isinstance(delta, dict):
                delta_content = delta.get("content")
                if isinstance(delta_content, str):
                    return delta_content
            content = output.get("content")
            if isinstance(content, str):
                return content

    if "choices" in data and isinstance(data.get("choices"), list) and data["choices"]:
        first_choice = data["choices"][0]
        if isinstance(first_choice, dict):
            delta = first_choice.get("delta")
            if isinstance(delta, dict):
                content = delta.get("content")
                if isinstance(content, str):
                    return content

    if "delta" in data and isinstance(data.get("delta"), dict):
        content = data["delta"].get("content")
        if isinstance(content, str):
            return content

    return None



from __future__ import annotations

from typing import Iterable, List, Optional

from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.utils.expression_factory import ExpressionFactory
from analogai.foundation.common.dtos.notion_expression import NotionExpression

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.features.question_answering.question_entity_extractor import QuestionEntityExtractor
from analogai.deepthink.features.question_answering.question_llm_caller import QuestionLlmCaller
from analogai.deepthink.features.question_answering.question_prompt_builder import (
    QuestionPromptBuilder,
    QuestionPromptContext,
    QuestionSynthesisContext,
)
from analogai.deepthink.infrastructure.gateways.llm.llm_gateway_provider_factory import (
    LlmGatewayProviderFactory,
)

try:
    from analogai.foundation.common.utils.expressions_serializer import serialize_belief_expression
except Exception:  # pragma: no cover - optional dependency
    serialize_belief_expression = None


class QuestionAnsweringService:
    def __init__(
        self,
        belief_query_service,
        notion_service,
        llm_gateway,
        synthesizer_gateway=None,
        entity_extractor: Optional[QuestionEntityExtractor] = None,
        prompt_builder: Optional[QuestionPromptBuilder] = None,
        llm_caller: Optional[QuestionLlmCaller] = None,
        synthesizer_caller: Optional[QuestionLlmCaller] = None,
        max_entity_depth: int = 5,
        similarity_cutoff: Optional[float] = None,
        use_vector_search: Optional[bool] = None,
    ):
        self._belief_query_service = belief_query_service
        self._notion_service = notion_service
        self._entity_extractor = entity_extractor or QuestionEntityExtractor()
        self._prompt_builder = prompt_builder or QuestionPromptBuilder()
        self._llm_caller = llm_caller or QuestionLlmCaller(llm_gateway)
        self._synthesizer_gateway = synthesizer_gateway
        self._synthesizer_caller = synthesizer_caller
        self._expression_factory = ExpressionFactory()
        self._max_entity_depth = max(0, int(max_entity_depth))
        self._similarity_cutoff = None
        if similarity_cutoff is not None:
            try:
                cutoff = float(similarity_cutoff)
                cutoff = max(0.0, min(1.0, cutoff))
                self._similarity_cutoff = cutoff
            except (TypeError, ValueError):
                self._similarity_cutoff = None
        self._use_vector_search = None
        if use_vector_search is not None:
            self._use_vector_search = bool(use_vector_search)

    def search(
        self,
        question_text: str,
        user_agent: UserAgent,
    ) -> Optional[str]:
        if not question_text or not question_text.strip():
            return None

        prompt_context, belief_count, sentence_count = self._build_prompt_context_with_counts(
            question_text=question_text,
            user_agent=user_agent,
        )
        if prompt_context is None:
            return None
        if belief_count + sentence_count == 0:
            return None
        user_message = self._prompt_builder.build_user_message(prompt_context)
        raw_answer = self._llm_caller.answer(user_message)

        synthesized = self._synthesize_answer(
            question_text=question_text.strip(),
            belief_context=prompt_context.belief_context,
            raw_answer=raw_answer,
        )
        return synthesized

    def generate_context(
        self,
        question_text: str,
        user_agent: UserAgent,
    ) -> str:
        if not question_text or not question_text.strip():
            return ""

        prompt_context, _, _ = self._build_prompt_context_with_counts(
            question_text=question_text,
            user_agent=user_agent,
        )
        if prompt_context is None:
            return ""
        return prompt_context.belief_context

    def get_context(
        self,
        question_text: str,
        user_agent: UserAgent,
    ) -> str:
        return self.generate_context(
            question_text=question_text,
            user_agent=user_agent,
        )

    def _synthesize_answer(
        self,
        question_text: str,
        belief_context: str,
        raw_answer: str,
    ) -> str:
        facts = _build_synth_facts(belief_context, raw_answer)
        if not facts:
            return ""
        synthesis_context = QuestionSynthesisContext(
            question=question_text,
            tasks="",
            facts=facts,
        )
        prompt = self._prompt_builder.build_synthesis_prompt(synthesis_context)
        caller = self._get_synthesizer_caller()
        return caller.answer(prompt)

    def _build_prompt_context_with_counts(
        self,
        question_text: str,
        user_agent: UserAgent,
    ) -> tuple[Optional[QuestionPromptContext], int, int]:
        entities = self._entity_extractor.extract(question_text)
        if not entities:
            return None, 0, 0
        beliefs = self._fetch_beliefs(user_agent, entities)
        notions = self._fetch_notions(user_agent, entities)
        notions = self._extend_notions_from_beliefs(user_agent, notions, beliefs)
        if not beliefs and not notions:
            return None, 0, 0
        belief_context = self._build_context_text(beliefs, notions)
        if not belief_context.strip():
            return None, 0, 0

        belief_count = self._count_serialized_beliefs(beliefs)
        sentence_count = self._count_context_lines(belief_context)

        return (
            QuestionPromptContext(
                question=question_text.strip(),
                entities=entities,
                belief_context=belief_context,
            ),
            belief_count,
            sentence_count,
        )

    def _count_serialized_beliefs(self, beliefs: Iterable) -> int:
        count = 0
        for belief in beliefs or []:
            if self._format_belief(belief):
                count += 1
        return count

    @staticmethod
    def _count_context_lines(text: str) -> int:
        return sum(1 for line in text.splitlines() if line.strip())
    def _get_synthesizer_caller(self) -> QuestionLlmCaller:
        if self._synthesizer_caller is not None:
            return self._synthesizer_caller
        if self._synthesizer_gateway is None:
            self._synthesizer_gateway = LlmGatewayProviderFactory.create_gateway(
                system_prompt="",
                max_tokens=200,
            )
        self._synthesizer_caller = QuestionLlmCaller(self._synthesizer_gateway)
        return self._synthesizer_caller

    def _fetch_beliefs(self, user_agent: UserAgent, entities: Iterable[str]) -> List:
        results: List = []
        seen = set()
        queue = [(entity, 0) for entity in entities if entity]
        seen_entities = {entity.lower() for entity, _ in queue}

        while queue:
            entity, depth = queue.pop(0)
            if depth > self._max_entity_depth:
                continue
            builder = self._belief_query_service.search(
                user_agent.user_id,
                user_agent.agent_id,
            ).where(
                Criterion.SubjectCaption, entity
            ).orWhere(
                Criterion.ComplementCaption, entity
            )
            self._apply_vector_search(builder)
            self._apply_similarity_cutoff(builder)
            try:
                beliefs = builder.list()
            except Exception:
                beliefs = []
            for belief in beliefs:
                key = getattr(belief, "id", None) or str(belief)
                if key in seen:
                    continue
                seen.add(key)
                results.append(belief)

                if depth < self._max_entity_depth:
                    for related in _extract_belief_entities(belief):
                        related_key = related.lower()
                        if related_key in seen_entities:
                            continue
                        seen_entities.add(related_key)
                        queue.append((related, depth + 1))

        return results

    def _apply_similarity_cutoff(self, builder) -> None:
        if self._similarity_cutoff is None:
            return
        for method_name in (
            "similarity_cutoff",
            "with_similarity_cutoff",
            "set_similarity_cutoff",
            "set_similarity_threshold",
            "similarity_threshold",
            "set_similarity_score_threshold",
            "set_min_similarity",
            "min_similarity",
        ):
            method = getattr(builder, method_name, None)
            if callable(method):
                method(self._similarity_cutoff)
                return

    def _apply_vector_search(self, builder) -> None:
        if self._use_vector_search is None:
            return
        for method_name in (
            "use_vector_search",
            "with_vector_search",
            "set_use_vector_search",
            "set_vector_search",
            "set_vector_search_enabled",
            "enable_vector_search",
            "vector_search",
            "use_vector",
        ):
            method = getattr(builder, method_name, None)
            if callable(method):
                try:
                    method(self._use_vector_search)
                except TypeError:
                    method()
                return

    def _format_beliefs(self, beliefs: Iterable) -> str:
        formatted = []
        for belief in beliefs:
            formatted.append(self._format_belief(belief))
        return "\n".join([line for line in formatted if line])

    def _format_notions(self, notions: Iterable) -> str:
        formatted: List[str] = []
        for notion in notions:
            try:
                formatted.append(str(notion))
            except Exception:
                continue
        return "\n".join([line for line in formatted if line])

    def _build_context_text(self, beliefs: Iterable, notions: Iterable) -> str:
        belief_text = self._format_beliefs(beliefs) if beliefs else ""
        notion_text = self._format_notions(notions) if notions else ""
        if belief_text and notion_text:
            return f"{belief_text}\n{notion_text}"
        return belief_text or notion_text

    def _format_belief(self, belief) -> str:
        try:
            return str(belief)
        except Exception:
            pass
        try:
            expression = self._expression_factory.propagate_belief_expression(belief)
            return _format_expression(expression)
        except Exception:
            return str(belief)

    def _fetch_notions(self, user_agent: UserAgent, entities: Iterable[str]) -> List:
        if self._notion_service is None:
            return []
        results: List = []
        seen = set()
        for entity in entities:
            if not entity:
                continue
            key = entity.strip().lower()
            if key in seen:
                continue
            seen.add(key)
            try:
                notion = self._notion_service.find(
                    user_agent.agent_id,
                    NotionExpression(caption=entity),
                )
            except Exception:
                notion = None
            if notion is not None:
                results.append(notion)
        return results

    def _extend_notions_from_beliefs(self, user_agent: UserAgent, notions: List, beliefs: Iterable) -> List:
        if self._notion_service is None:
            return notions
        results = list(notions) if notions else []
        seen = {getattr(notion, "id", None) or getattr(notion, "caption", "") for notion in results}
        for belief in beliefs or []:
            for notion in (getattr(belief, "subject_notion", None), getattr(belief, "complement_notion", None)):
                if notion is None:
                    continue
                notion_id = getattr(notion, "id", None)
                notion_caption = getattr(notion, "caption", None)
                key = notion_id or notion_caption or str(notion)
                if key in seen:
                    continue
                if notion_caption:
                    try:
                        fetched = self._notion_service.find(
                            user_agent.agent_id,
                            NotionExpression(caption=notion_caption),
                        )
                    except Exception:
                        fetched = None
                    if fetched is not None:
                        results.append(fetched)
                        seen.add(getattr(fetched, "id", None) or getattr(fetched, "caption", "") or key)
                        continue
                results.append(notion)
                seen.add(key)
        return results


def _format_expression(expression) -> str:
    subject_expr = getattr(expression, "subject_notion_expression", None) or getattr(expression, "subject", None)
    complement_expr = getattr(expression, "complement_notion_expression", None) or getattr(expression, "complement", None)
    relation = getattr(expression, "relation", None)
    probability = getattr(expression, "probability", None)
    modifier = getattr(expression, "modifier", None)

    subject = _caption(subject_expr)
    complement = _caption(complement_expr)
    relation_value = getattr(relation, "value", None) or str(relation) if relation is not None else ""

    parts = [p for p in [subject, relation_value, complement] if p]
    base = " ".join(parts) if parts else str(expression)

    modifier_bits = []
    if modifier is not None:
        location = getattr(modifier, "location", None)
        time_value = getattr(modifier, "time", None)
        quantity = getattr(modifier, "quantity", None)
        deontic = getattr(modifier, "deontic_modality", None)
        if location:
            modifier_bits.append(f"location={location}")
        if time_value:
            modifier_bits.append(f"time={time_value}")
        if quantity is not None:
            modifier_bits.append(f"quantity={quantity}")
        if deontic:
            modifier_bits.append(f"deontic={deontic}")

    if probability is not None:
        modifier_bits.append(f"probability={probability}")

    if modifier_bits:
        return f"{base} ({', '.join(modifier_bits)})"
    return base


def _caption(value) -> str:
    if value is None:
        return ""
    caption = getattr(value, "caption", None)
    attributes = _extract_notion_attributes(value)
    if caption:
        if attributes:
            return f"{caption} ({_format_attributes(attributes)})"
        return str(caption)
    if attributes:
        return _format_attributes(attributes)
    return str(value)


def _extract_belief_entities(belief) -> List[str]:
    entities = []
    subject_notion = getattr(belief, "subject_notion", None)
    complement_notion = getattr(belief, "complement_notion", None)
    if subject_notion is not None:
        caption = getattr(subject_notion, "caption", None)
        if caption:
            entities.append(str(caption))
    if complement_notion is not None:
        caption = getattr(complement_notion, "caption", None)
        if caption:
            entities.append(str(caption))
    return entities


def _extract_notion_attributes(notion) -> list:
    if notion is None:
        return []
    attributes = getattr(notion, "attributes", None)
    if attributes is None:
        return []
    if isinstance(attributes, list):
        return attributes
    return list(attributes) if attributes else []


def _format_notion_attributes(notion) -> str:
    caption = getattr(notion, "caption", None)
    attributes = _extract_notion_attributes(notion)
    if not attributes:
        return ""
    attrs_text = _format_attributes(attributes)
    if caption:
        return f"{caption} ({attrs_text})"
    return attrs_text


def _format_attributes(attributes: list) -> str:
    parts: List[str] = []
    for attr in attributes:
        tag = getattr(attr, "tag", None) or (attr.get("tag") if isinstance(attr, dict) else None)
        value = getattr(attr, "value", None) if not isinstance(attr, dict) else attr.get("value")
        unit = getattr(attr, "unit", None) if not isinstance(attr, dict) else attr.get("unit")

        if value is None and hasattr(attr, "values"):
            values = getattr(attr, "values", None)
            if values:
                value = values

        if value is None:
            if tag:
                parts.append(str(tag))
            continue

        if isinstance(value, list):
            value_str = ", ".join(str(v) for v in value)
        else:
            value_str = str(value)

        unit_str = f" {unit}" if unit else ""
        if tag:
            parts.append(f"{tag}={value_str}{unit_str}")
        else:
            parts.append(f"{value_str}{unit_str}")
    return ", ".join(parts)


def _build_synth_facts(belief_context: str, raw_answer: str) -> str:
    lines: List[str] = []
    if belief_context:
        for line in belief_context.splitlines():
            trimmed = line.strip()
            if trimmed:
                lines.append(f"- [belief] {trimmed}")
    if raw_answer and raw_answer.strip():
        lines.append(f"- [qa] {raw_answer.strip()}")
    return "\n".join(lines)



from __future__ import annotations

import json
import re
from typing import Iterable, List, Optional

from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.infrastructure.gateways.llm.llm_factory import LlmGatewayFactory
from analogai.deepthink.infrastructure.gateways.llm.llm_gateway_provider_factory import LlmGatewayProviderFactory


_DEFAULT_STOPWORDS = {
    "a", "an", "and", "are", "as", "at", "be", "by", "can", "could", "did", "do", "does",
    "for", "from", "had", "has", "have", "how", "i", "if", "in", "is", "it", "may", "might",
    "of", "on", "or", "should", "so", "that", "the", "their", "them", "then", "there", "these",
    "they", "this", "to", "was", "were", "what", "when", "where", "which", "who", "why", "will",
    "with", "would", "you", "your",
    "sentences", "intent_type", "intent", "subject", "predicate", "relation", "object", "complement",
    "wrestler", "show", "event", "events", "match", "matches", "tournament",
    "film", "films", "movie", "movies", "show", "series", "episode", "episodes",
}


class QuestionEntityExtractor:
    def __init__(self, llm_gateway=None, max_llm_tokens: int = 600):
        self._llm_gateway = llm_gateway
        self._max_llm_tokens = max_llm_tokens

    def extract(self, text: str, max_entities: int = 8) -> List[str]:
        if not text:
            return []
        candidates = []
        candidates.extend(self._extract_with_llm(text))
        candidates.extend(_fallback_tokens(text))

        normalized = _dedupe_preserve_order(_normalize_entities(candidates))
        return normalized[:max_entities]

    def _extract_with_llm(self, text: str) -> List[str]:
        gateway = self._llm_gateway or LlmGatewayProviderFactory.create_gateway(
            system_prompt="",
            max_tokens=self._max_llm_tokens,
        )
        sanitized = text.replace('"', "")
        prompts = [
            (
                "You extract named entities from a question. Output ONLY valid JSON (no markdown, no prose). "
                "Schema: {\"entities\": [\"...\", \"...\"]}. Use double quotes; no trailing commas. "
                "\n\nTHINKING (do not output):\n"
                "1) Identify only proper names and titled works.\n"
                "2) Exclude numbers, years, dates, ages, quantities.\n"
                "3) Exclude roles/titles/occupations and generic nouns (e.g., wrestler, director, show, album, movie, event, city, country).\n"
                "4) Keep only the named entity strings.\n\n"
                "DO NOT return: roles, descriptors, generic nouns, adjectives, or numeric tokens.\n"
                "DO NOT include: subjects/relations/predicates/certainty or any extra fields.\n\n"
                "Examples:\n"
                "Q: Which wrestler, who died in 2009, featured on the wrestling show Starrcade?\n"
                "A: {\"entities\": [\"Starrcade\"]}\n"
                "Q: The song Medicine by The 1975 appears in a film directed by what Danish filmmaker?\n"
                "A: {\"entities\": [\"Medicine\", \"The 1975\"]}\n"
                "Q: World for Ransom was directed by the producer notable for which 1974 film?\n"
                "A: {\"entities\": [\"World for Ransom\"]}\n\n"
                f"Sentence: {sanitized}"
            ),
            (
                "Return ONLY valid JSON with double quotes and no trailing commas. "
                "Schema: {\"entities\": [\"...\", \"...\"]}. "
                "Only named entities (proper nouns / titled works). "
                "Exclude years, dates, numbers, roles/titles/occupations, and generic nouns. "
                "Examples: 'Danish filmmaker' -> exclude; '1974' -> exclude; 'director' -> exclude. "
                f"Sentence: {sanitized}"
            ),
        ]

        last_response = ""
        for attempt, user_message in enumerate(prompts, start=1):
            try:
                last_response = gateway.generate_completion(user_message)
            except Exception as exc:
                app_logger.warning(
                    "[QuestionEntityExtractor] LLM extraction failed. "
                    f"attempt={attempt} reason={exc} text={text}"
                )
                continue

            try:
                data = _safe_json_loads(last_response)
                return _extract_entities_from_json(data)
            except json.JSONDecodeError as exc:
                app_logger.warning(
                    "[QuestionEntityExtractor] Invalid JSON from LLM. "
                    f"attempt={attempt} reason={exc} text={text} llm_output={last_response}"
                )

        return []


def _fallback_tokens(text: str) -> Iterable[str]:
    words = re.findall(r"[A-Za-z][A-Za-z0-9'_-]*", text)
    tokens: List[str] = []
    for word in words:
        lower = word.lower()
        if lower in _DEFAULT_STOPWORDS:
            continue
        if len(lower) <= 2:
            continue
        tokens.append(word)
    tokens.extend(_extract_years(text))
    return tokens


def _extract_years(text: str) -> List[str]:
    years = re.findall(r"\b(18\d{2}|19\d{2}|20\d{2})\b", text)
    return list(dict.fromkeys(years))


def _normalize_entities(entities: Iterable[str]) -> List[str]:
    normalized: List[str] = []
    for entity in entities:
        cleaned = re.sub(r"\s+", " ", entity).strip(" \t\n\r\f\v.,!?;:\"'()[]{}")
        if not cleaned:
            continue
        normalized.append(cleaned)
    return normalized


def _dedupe_preserve_order(items: Iterable[str]) -> List[str]:
    seen = set()
    result = []
    for item in items:
        key = item.lower()
        if key in seen:
            continue
        seen.add(key)
        result.append(item)
    return result


def _extract_entities_from_json(data) -> List[str]:
    if not isinstance(data, dict):
        return []
    entities_field = data.get("entities")
    if isinstance(entities_field, list):
        entities: List[str] = []
        for item in entities_field:
            if isinstance(item, str):
                entities.append(item)
        return entities
    sentences = data.get("sentences")
    if isinstance(sentences, list):
        items = []
        for sentence in sentences:
            items.extend(_extract_entities_from_json(sentence))
        return items

    entities: List[str] = []
    entities.extend(_extract_entities_from_node(data.get("subject")))
    predicate = data.get("predicate")
    if isinstance(predicate, dict):
        entities.extend(_extract_entities_from_node(predicate.get("complement")))

    return entities


def _safe_json_loads(text: str):
    cleaned = text.strip()
    if not cleaned:
        raise json.JSONDecodeError("Empty JSON", cleaned, 0)

    if cleaned.startswith("```"):
        cleaned = re.sub(r"^```[a-zA-Z]*\n?", "", cleaned)
        cleaned = re.sub(r"```$", "", cleaned).strip()

    try:
        return json.loads(cleaned)
    except json.JSONDecodeError:
        pass

    start = cleaned.find("{")
    end = cleaned.rfind("}")
    if start != -1 and end != -1 and end > start:
        snippet = cleaned[start:end + 1]
        return json.loads(snippet)

    return json.loads(cleaned)


def _extract_entities_from_node(node: Optional[object]) -> List[str]:
    if node is None:
        return []
    if isinstance(node, str):
        return [node]
    if isinstance(node, dict):
        caption = node.get("caption")
        entities = [caption] if isinstance(caption, str) else []
        attributes = node.get("attributes")
        if isinstance(attributes, list):
            for attr in attributes:
                if not isinstance(attr, dict):
                    continue
                value = attr.get("value")
                if isinstance(value, str):
                    entities.append(value)
        return entities
    return []



import unittest
import uuid

from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.tests.e2e.base_deepthink_e2e_test import BaseThinkerE2ETest


class TestMultiUserThinkerE2E(BaseThinkerE2ETest):
    def setUp(self):
        super().setUp()
        self._user1_id = str(uuid.uuid4())
        self._user2_id = str(uuid.uuid4())
        self._agent1_id = str(uuid.uuid4())
        self._agent2_id = str(uuid.uuid4())

    def _think_user1_agent1(self, message: str):
        return self._deepthink.deepthink(
            message,
            user_id=self._user1_id,
            agent_id=self._agent1_id,
            previous_messages_stack=PreviousMessagesStack(),
        )

    def _think_user2_agent1(self, message: str):
        return self._deepthink.deepthink(
            message,
            user_id=self._user2_id,
            agent_id=self._agent1_id,
            previous_messages_stack=PreviousMessagesStack(),
        )

    def _think_user1_agent2(self, message: str):
        return self._deepthink.deepthink(
            message,
            user_id=self._user1_id,
            agent_id=self._agent2_id,
            previous_messages_stack=PreviousMessagesStack(),
        )

    def test_knowledge_persists_across_different_users_same_agent(self):
        self._think_user1_agent1("socrates is human")
        response = self._think_user2_agent1("is socrates human?")
        self.assertIsNotNone(response)
        message_text = (response or "").lower()
        self.assertTrue(
            "socrates" in message_text or "human" in message_text,
            f"Knowledge should persist across users with same agent. Got: {message_text[:200]}",
        )

    def test_user_isolation(self):
        self._think_user1_agent1("I like apples")
        response = self._think_user2_agent1("What do I like?")
        self.assertIsNone(response)

    def test_agent_knowledge_isolation(self):
        self._think_user1_agent1("Philosophy is interesting")
        response = self._think_user1_agent2("What is interesting?")
        self.assertIsNone(response)

    def test_multi_agent_conversation_flow(self):
        response1 = self._think_user1_agent1("Hello, I'm Alice")
        response2 = self._think_user2_agent1("Hello, I'm Bob")
        self.assertIsNotNone(response1)
        self.assertIsNotNone(response2)

    def test_concurrent_multi_user_multi_agent_scenario(self):
        self._think_user1_agent1("User 1 message to agent 1")
        self._think_user2_agent1("User 2 message to agent 2")
        self._think_user1_agent1("Another message from user 1")
        self._think_user2_agent1("Another message from user 2")


if __name__ == "__main__":
    unittest.main()



from analogai.foundation.common.dtos.belief_expression import BeliefExpression
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.deepthink.application.usecases.learning.make_direct_statement.models import MakeDirectStatementResponse
from analogai.foundation.common.utils.expression_factory import ExpressionFactory


class LearnerResponseBuilder:
    def __init__(self, context):
        self.context = context
        self._expression_factory = ExpressionFactory()

    def build_response(self, probability=None):
        if self.context.statement:
            belief_expression = self._expression_factory.propagate_belief_expression(self.context.statement)
        else:
            subject_notion_expression = self._express_notion(self.context.subject_notion) if self.context.subject_notion else self.context.subject_notion_expression
            complement_notion_expression = self._express_notion(self.context.complement_notion) if self.context.complement_notion else self.context.complement_notion_expression
            if probability is not None:
                response_probability = probability
            else:
                response_probability = self.context.proposed_probability

            relation = self.context.relation
            belief_expression = BeliefExpression(
                subject_notion_expression=subject_notion_expression,
                complement_notion_expression=complement_notion_expression,
                relation=relation,
                probability=response_probability,
                validity=1.0,
                modifier=self.context.modifier,
            )

        return MakeDirectStatementResponse(
            user_agent=self.context.user_agent,
            belief_expression=belief_expression
        )

    def _express_notion(self, notion):
        from analogai.deepthink.application.utils.notion_expresser_util import generate_expression
        return generate_expression(notion)




from typing import List
import logging

from .propagated_sentence import PropagatedSentence
from .propagated_sentence_builder import build_propagated_sentence
from .json_sentence_parser import parse_json, is_json_format


logger = logging.getLogger(__name__)


class MessageParser:
    def parse(self, text: str) -> List[PropagatedSentence]:
        if not text or not text.strip():
            return []
        text = text.strip()
        if not is_json_format(text):
            return []
        
        logger.debug(f"[1] Parsing JSON: {text[:100]}...")
        parsed_sentences = parse_json(text)
        logger.debug(f"[2] Extracted {len(parsed_sentences)} sentences: {parsed_sentences}")
        
        result = [build_propagated_sentence(s) for s in parsed_sentences]
        logger.debug(f"[3] Enum propagated: {result}")
        
        return result



from __future__ import annotations

from typing import Callable, Optional

from analogai.deepthink.application.value_objects.user_agent import UserAgent
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.features.question_answering.question_answering_service import QuestionAnsweringService
from analogai.deepthink.features.question_answering.question_detection import is_question


class QuestionRouter:
    def __init__(self, question_service: QuestionAnsweringService):
        self._question_service = question_service

    def is_question(self, text: str) -> bool:
        return is_question(text)

    def route(
        self,
        raw_message: str,
        user_agent: UserAgent,
        previous_messages_stack: PreviousMessagesStack,
        statement_handler: Callable[[str, UserAgent, PreviousMessagesStack], Optional[str]],
    ) -> Optional[str]:
        if self.is_question(raw_message):
            return self._question_service.search(
                question_text=raw_message,
                user_agent=user_agent,
            )
        return statement_handler(raw_message, user_agent, previous_messages_stack)



import unittest
from analogai.deepthink.presentation.postprocessing.postprocessor import postprocess_text
from analogai.deepthink.presentation.postprocessing.postprocessors.pronoun_postprocessior import postprocesses_pronouns
from analogai.deepthink.presentation.enums.pronoun_placeholder import PronounPlaceholder


class TestPostprocessorIntegration(unittest.TestCase):
    def test_postprocessor_chain(self):
        input_text = f"{PronounPlaceholder.AGENT_PLACEHOLDER.value} is human friend"
        postprocessed = postprocess_text(input_text)
        self.assertIsNotNone(postprocessed)
        self.assertIsInstance(postprocessed, str)
        self.assertEqual(postprocessed, "I am human friend")

    def test_postprocessor_with_pronouns(self):
        input_text = f"{PronounPlaceholder.AGENT_PLACEHOLDER.value} is human friend"
        postprocessed = postprocess_text(input_text)
        self.assertIsNotNone(postprocessed)
        self.assertIsInstance(postprocessed, str)
        self.assertEqual(postprocessed, "I am human friend")

    def test_postprocessor_with_sentence(self):
        input_text = f"is {PronounPlaceholder.AGENT_PLACEHOLDER.value} human friend"
        postprocessed = postprocess_text(input_text)
        self.assertIsNotNone(postprocessed)
        self.assertIsInstance(postprocessed, str)

    def test_postprocessor_with_multiple_pronouns(self):
        agent = PronounPlaceholder.AGENT_PLACEHOLDER.value
        input_text = f"{agent} could be human friend and {agent} is smart"
        postprocessed = postprocess_text(input_text)
        self.assertIsNotNone(postprocessed)
        self.assertIsInstance(postprocessed, str)
        self.assertEqual(postprocessed, "I could be human friend and I am smart")

    def test_pronoun_replacement_after_grammar_correction(self):
        respondent = PronounPlaceholder.RESPONDENT_PLACEHOLDER.value
        grammar_corrected = f"{respondent} can sell the iphone 12 today"
        pronoun_formatted = postprocesses_pronouns(grammar_corrected)
        self.assertIn("you", pronoun_formatted.lower())
        self.assertIn("the iphone 12", pronoun_formatted.lower())
        self.assertNotIn(respondent, pronoun_formatted)


if __name__ == "__main__":
    unittest.main()





from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from analogai.deepthink.startup.factories.domain_service_factory import DomainServiceFactory


class ExtensionsServiceFactory:
    def __init__(
        self,
        domain_service_factory: "DomainServiceFactory",
        env=None,
        config: Optional[object] = None,
    ):
        self._domain_service_factory = domain_service_factory
        self._env = env
        self._config = config
        self._conversation_service = None
        self._authorization_factory = None


    def get_conversation_service(self):
        if self._conversation_service is None:
            from analogai.foundation.extensions.conversation.factories.conversation_service_factory import (
                ConversationServiceFactory,
            )

            self._conversation_service = ConversationServiceFactory().create()
        return self._conversation_service


    def get_authorization_factory(self):
        if self._authorization_factory is None:
            from analogai.foundation.extensions.authorization.authorization_factory import AuthorizationFactory

            self._authorization_factory = AuthorizationFactory()
        return self._authorization_factory


    def get_belief_query_service(self):
        return _create_belief_query_service(self._domain_service_factory)

    def get_statement_derivation_service(self):
        return _create_statement_derivation_service(self._domain_service_factory)


def _create_belief_query_service(domain_service_factory: "DomainServiceFactory"):
    belief_service = domain_service_factory.get_belief_service()
    from analogai.foundation.setup.factories.belief_query_service_factory import (
        BeliefQueryServiceFactory,
    )

    return BeliefQueryServiceFactory(
        belief_service=belief_service,
    ).get_service()


def _create_statement_derivation_service(domain_service_factory: "DomainServiceFactory"):
    return domain_service_factory.get_direct_statement_service()



from unittest.mock import Mock
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.enums import RelationshipType

class MockRequestFactory:
    request_class = None

    @classmethod
    def create(cls, subject_notion_expression=None, complement_notion_expression=None, relationship_type=RelationshipType.IS_A, **kwargs):
        if subject_notion_expression is None:
            subject_notion_expression = NotionExpression(caption="test_subject")
        if complement_notion_expression is None:
            complement_notion_expression = NotionExpression(caption="test_object")
        request = Mock(spec=cls.request_class)
        request.subject_notion_expression = subject_notion_expression
        request.complement_notion_expression = complement_notion_expression
        request.relationship_type = relationship_type
        for key, value in kwargs.items():
            setattr(request, key, value)
        return request 





import json
import re
from pathlib import Path

from analogai.deepthink.infrastructure.gateways.llm.llm_factory import LlmGatewayFactory
from analogai.deepthink.infrastructure.gateways.llm.providers.sentence_transformer_examples import (
    get_enhanced_system_prompt,
)
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.deepthink.presentation.preprocessing.previous_messages_stack import PreviousMessagesStack
from analogai.deepthink.presentation.parsers.json_sentence_parser import is_json_format


def get_sentence_transformer():
    return transform_sentence_with_llm


def _normalize_to_sentences_list(data: dict) -> dict:
    if "sentences" in data:
        return data
    if "relationships" in data:
        return {"sentences": data["relationships"]}
    if "intent" in data or "intent_type" in data:
        return {"sentences": [data]}
    return data


def _safe_json_loads(text: str):
    cleaned = text.strip()
    if not cleaned:
        raise json.JSONDecodeError("Empty JSON", cleaned, 0)

    if cleaned.startswith("```"):
        cleaned = re.sub(r"^```[a-zA-Z]*\n?", "", cleaned)
        cleaned = re.sub(r"```$", "", cleaned).strip()

    try:
        return json.loads(cleaned)
    except json.JSONDecodeError:
        repaired = _repair_sentence_array_json(cleaned)
        if repaired != cleaned:
            try:
                return json.loads(repaired)
            except json.JSONDecodeError:
                pass

    decoder = json.JSONDecoder()
    index = 0
    length = len(cleaned)
    objects = []

    while index < length:
        next_obj = cleaned.find("{", index)
        next_arr = cleaned.find("[", index)
        if next_obj == -1 and next_arr == -1:
            break
        if next_obj == -1 or (next_arr != -1 and next_arr < next_obj):
            index = next_arr
        else:
            index = next_obj

        try:
            obj, end = decoder.raw_decode(cleaned[index:])
        except json.JSONDecodeError:
            break

        objects.append(obj)
        index += end

        while index < length and cleaned[index] in " \t\r\n,":
            index += 1

    if objects:
        if len(objects) == 1:
            return objects[0]
        if all(isinstance(item, dict) for item in objects):
            return {"sentences": objects}
        return objects

    return json.loads(cleaned)


def _repair_sentence_array_json(text: str) -> str:
    if '"sentences"' not in text:
        return text
    pattern = re.compile(r'(\]\}\}|\}\})\s*,\s*\{"intent_type"')

    def _insert_missing_brace(match: re.Match) -> str:
        return f"{match.group(1)}}},{{\"intent_type\""

    repaired = pattern.sub(_insert_missing_brace, text)
    return repaired


def _load_sentence_transformer_prompt() -> str:
    prompts_dir = Path(__file__).resolve().parents[3] / "infrastructure" / "prompts"
    prompt_path = prompts_dir / "sentence_transformer_prompt.txt"
    if not prompt_path.exists():
        raise FileNotFoundError(f"Prompt file not found: {prompt_path}")
    base_prompt = prompt_path.read_text(encoding="utf-8").strip()
    return get_enhanced_system_prompt(base_prompt)


def _format_json_for_log(value: str) -> str:
    if value is None:
        return "(none)"
    text = value.strip()
    if not text:
        return "(empty)"
    try:
        parsed = _safe_json_loads(text)
        return json.dumps(parsed, indent=2, ensure_ascii=False)
    except json.JSONDecodeError:
        return text


def transform_sentence_with_llm(text: str, previous_messages_stack: PreviousMessagesStack) -> str:
    if not text or not text.strip():
        return text

    if is_json_format(text):
        app_logger.info(f"[SentenceTransformer] Skipping LLM - already JSON format")
        try:
            data = json.loads(text.strip())
            normalized = _normalize_to_sentences_list(data)
            return json.dumps(normalized)
        except json.JSONDecodeError as e:
            app_logger.error(
                "[SentenceTransformer] Invalid JSON input provided. "
                f"reason={e} input_text={text}"
            )
            return text

    try:
        gateway = LlmGatewayFactory.create_sentence_decomposer_and_transformer()

        sanitized_text = text.replace('"', "").strip()
        if sanitized_text.startswith(("- ", "* ", "• ")):
            sanitized_text = sanitized_text[2:].strip()
        sanitized_text = re.sub(r"^\d+[\.)]\s+", "", sanitized_text)

        history_pairs = previous_messages_stack.get_history_pairs()
        history = []
        for role, content in history_pairs:
            if not content or not content.strip():
                continue
            label = "User" if role == "user" else "Agent"
            history.append(f"{label}: {content.strip()}")
        instruction = _load_sentence_transformer_prompt()
        if hasattr(gateway, "set_system_prompt"):
            gateway.set_system_prompt(instruction)

        if history:
            history_text = "\n".join(history)
            user_message = (
                f"Conversation history (for context only): {history_text}\n\n"
                f"Sentence: {sanitized_text}"
            )
        else:
            user_message = f"Sentence: {sanitized_text}"

        app_logger.info(
            "[SentenceTransformer] LLM call\n"
            f"Input:\n{user_message}\n"
        )
        transformed = gateway.generate_completion(user_message)
        app_logger.info(
            "[SentenceTransformer] LLM output\n"
            f"Output:\n{_format_json_for_log(transformed)}\n"
        )
        try:
            data = _safe_json_loads(transformed)
        except json.JSONDecodeError as e:
            app_logger.warning(
                "[SentenceTransformer] Invalid JSON from LLM, retrying once. "
                f"reason={e} input_text={text}"
            )
            retry_transformed = gateway.generate_completion(user_message)
            app_logger.info(
                "[SentenceTransformer] LLM retry output\n"
                f"Output:\n{_format_json_for_log(retry_transformed)}\n"
            )
            try:
                data = _safe_json_loads(retry_transformed)
                transformed = retry_transformed
            except json.JSONDecodeError as retry_error:
                app_logger.error(
                    "[SentenceTransformer] LLM returned invalid JSON. "
                    f"reason={retry_error} input_text={text} history={history} "
                    f"user_message={user_message} llm_output={retry_transformed}"
                )
                return text
        normalized = _normalize_to_sentences_list(data)

        return json.dumps(normalized)
    except Exception as e:
        app_logger.error(
            "[SentenceTransformer] Error transforming sentence with LLM. "
            f"reason={e} input_text={text}"
        )
        return text



from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.dtos.notion_expression import NotionExpression

from analogai.deepthink.application.usecases.learning.make_notion_statement.models import (
    MakeNotionStatementRequest,
    MakeNotionStatementResponse,
)
from analogai.deepthink.application.usecases.learning.make_notion_statement.ports import (
    MakeNotionStatementOutputPort,
)


class MakeNotionStatementUseCase:
    def __init__(self, notion_service: NotionService, output_port: MakeNotionStatementOutputPort):
        self._notion_service = notion_service
        self._output_port = output_port

    def execute(self, request: MakeNotionStatementRequest):
        if request is None or request.notion_expression is None:
            return None

        notion_expression = request.notion_expression
        if not isinstance(notion_expression, NotionExpression):
            return None

        app_logger.info(
            "Saving notion statement: caption='%s' attributes=%s",
            notion_expression.caption,
            notion_expression.attributes,
        )

        self._notion_service.upsert_from_expression(
            request.user_agent.user_id,
            request.user_agent.agent_id,
            notion_expression,
        )

        response = MakeNotionStatementResponse(
            user_agent=request.user_agent,
            notion_expression=notion_expression,
            saved=True,
        )
        return self._output_port.output_notion_statement_saved(response)



from dotenv import load_dotenv
import os

load_dotenv()

OPENAI_API_KEY = os.getenv('OPENAI_API_KEY', '')
ANTHROPIC_API_KEY = os.getenv('ANTHROPIC_API_KEY', '')
EMBEDDING_MODEL_NAME = os.getenv('EMBEDDING_MODEL_NAME', 'text-embedding-3-small')

XAI_API_KEY = os.getenv('XAI_API_KEY', '')

AWS_ACCESS_KEY_ID = os.getenv('AWS_ACCESS_KEY_ID', '')
AWS_SECRET_ACCESS_KEY = os.getenv('AWS_SECRET_ACCESS_KEY', '')
AWS_REGION = os.getenv('AWS_REGION', 'us-east-1')
BEDROCK_MODEL_ID = os.getenv('BEDROCK_MODEL_ID', 'anthropic.claude-3-sonnet-20240229-v1:0')

AZURE_AI_FOUNDRY_ENDPOINT = os.getenv('AZURE_AI_FOUNDRY_ENDPOINT', '')
AZURE_AI_FOUNDRY_API_KEY = os.getenv('AZURE_AI_FOUNDRY_API_KEY', '')

AZURE_OPENAI_ENDPOINT = os.getenv('AZURE_OPENAI_ENDPOINT', '')
AZURE_OPENAI_API_KEY = os.getenv('AZURE_OPENAI_API_KEY', '')

TRACK_LLM = os.getenv('TRACK_LLM', 'true').lower() == 'true'

OLLAMA_HOST = os.getenv('OLLAMA_HOST', 'localhost')
OLLAMA_PORT = os.getenv('OLLAMA_PORT', '11434')
OLLAMA_MODEL_NAME = os.getenv('OLLAMA_MODEL_NAME', 'qwen2.5:0.5b-instruct')



from analogai.foundation.common.logging.app_logger import app_logger

def normalize_also_separated(text: str) -> str:
    if not text or not text.strip():
        return text
    
    if '[also]' in text:
        parts = [part.strip() for part in text.split('[also]') if part.strip()]
        if len(parts) > 1:
            normalized = '[also]'.join(parts)
            app_logger.info(f"Normalized [also] separated text: {len(parts)} relationships")
            return normalized
        return text
    
    lines = [line.strip() for line in text.split('\n') if line.strip()]
    if len(lines) > 1:
        normalized = '[also]'.join(lines)
        app_logger.info(f"Normalized newline separated text: converted {len(lines)} lines to [also] format")
        return normalized
    
    return text




from typing import Iterator, Optional, Tuple

from analogai.foundation.common.enums import RelationshipType


def _normalize_tag_key(s: str) -> str:
    return s.strip().lower().replace(" ", "_")


def tag_key_to_relationship_type(tag_key: str) -> Optional[RelationshipType]:
    normalized = _normalize_tag_key(tag_key)
    if normalized.startswith("is_") and normalized not in ("is", "is_a", "isa"):
        normalized = normalized[3:]
    for rt in RelationshipType:
        if rt.name.lower() == normalized:
            return rt
        if _normalize_tag_key(rt.value) == normalized:
            return rt
    return None


def relationship_type_tag_keys() -> Iterator[Tuple[str, RelationshipType]]:
    for rt in RelationshipType:
        yield rt.value, rt
        key_from_name = rt.name.lower()
        if key_from_name != rt.value:
            yield key_from_name, rt



from typing import cast, TYPE_CHECKING
from analogai.deepthink.presentation.interpreters.interpreter_result import InterpreterResult

if TYPE_CHECKING:
    from analogai.deepthink.presentation.factories.interpreters_factory import InterpretersFactory


_NO_RESULT = object()


class InterpreterPipeline:
    def __init__(self, interpreters_factory: "InterpretersFactory"):
        self._interpreters_factory = interpreters_factory
        self._interpreters = None
        self._result_rules = (
            (self._is_false_result, self._return_false),
            (self._has_value_result, self._return_value),
        )

    def _get_interpreters(self):
        if self._interpreters is None:
            self._interpreters = self._interpreters_factory.create_all()
        return self._interpreters

    def _resolve_interpreter_result(self, result):
        for predicate, mapper in self._result_rules:
            if predicate(result):
                return mapper(result)
        return _NO_RESULT

    @staticmethod
    def _is_false_result(result) -> bool:
        return result is False

    @staticmethod
    def _has_value_result(result) -> bool:
        return result is not None

    @staticmethod
    def _return_false(result):
        return False

    @staticmethod
    def _return_value(result):
        return result

    def run(self, parsed_message, user_agent, message_text: str) -> InterpreterResult | None | bool:
        for interpreter in self._get_interpreters():
            result = interpreter.interpret(parsed_message, user_agent, message_text)
            resolved = self._resolve_interpreter_result(result)
            if resolved is not _NO_RESULT:
                return cast(InterpreterResult | bool | None, resolved)
        return None


if __name__ == "__main__":
    from analogai.deepthink.application.value_objects.user_agent import UserAgent
    from analogai.deepthink.presentation.parsers.parsers_factory import build_message_parser
    pipeline = InterpreterPipeline(InterpretersFactory())
    parser = build_message_parser()
    text = [
            "Iphone X has USB-C port",
            "iphone x has usb-c port",
            "John the programmer",
            "Maria, a talented doctor",
            "my boss Sarah the perfectionist",
            "Elon Musk, founder of SpaceX",
            "I mean the programmer",
            "I mean John, the programmer"]
    for t in text:
        parsed_list = parser.parse(t)
        parsed = parsed_list[0] if parsed_list else None
        result = pipeline.run(parsed, UserAgent(user_id="u", agent_id="a", user_authority=1.0), t)
        print(result)
        print("--------------------------------")
